#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import os
import argparse

from migen import *

from litex.build.generic_platform import Pins, Subsignal
from litex.build.sim.config import SimConfig

from litex.soc.interconnect.csr import CSR
from litex.soc.integration.soc_core import SoCCore, soc_core_args, soc_core_argdict
from litex.soc.integration.builder import builder_args, builder_argdict, Builder
from litex.soc.cores.cpu import CPUS

from litedram.gen import LiteDRAMCoreControl
from litedram import modules as litedram_modules
from litedram.core.controller import ControllerSettings
from litedram.phy.model import DFITimingsChecker, _speedgrade_timings, _technology_timings

from litedram.phy.lpddr4.simphy import LPDDR4SimPHY, DoubleRateLPDDR4SimPHY
from litedram.phy.lpddr4.sim import LPDDR4Sim

from litedram.phy.sim_utils import Clocks, CRG, Platform

# Platform -----------------------------------------------------------------------------------------

_io = [
    # clocks added in main()
    ("lpddr4", 0,
        Subsignal("clk",   Pins(1)),
        # Subsignal("clk_n",   Pins(1)),
        Subsignal("cke",     Pins(1)),
        Subsignal("odt",     Pins(1)),
        Subsignal("reset_n", Pins(1)),
        Subsignal("cs",      Pins(1)),
        Subsignal("ca",      Pins(6)),
        Subsignal("dqs",     Pins(2)),
        # Subsignal("dqs_n",   Pins(2)),
        Subsignal("dmi",     Pins(2)),
        Subsignal("dq",      Pins(16)),
    ),
]

# Clocks -------------------------------------------------------------------------------------------

def get_clocks(sys_clk_freq):
    return Clocks({
        "sys":           dict(freq_hz=sys_clk_freq),
        "sys_11_25":     dict(freq_hz=sys_clk_freq, phase_deg=11.25),
        "sys2x":         dict(freq_hz=2*sys_clk_freq),
        "sys8x":         dict(freq_hz=8*sys_clk_freq),
        "sys8x_ddr":     dict(freq_hz=2*8*sys_clk_freq),
        "sys8x_90":      dict(freq_hz=8*sys_clk_freq, phase_deg=90),
        "sys8x_90_ddr":  dict(freq_hz=2*8*sys_clk_freq, phase_deg=2*90),
    })

# SoC ----------------------------------------------------------------------------------------------

class SimSoC(SoCCore):
    """Simulation of SoC with LPDDR4 DRAM

    This is a SoC used to run Verilator-based simulations of LiteDRAM with a simulated LPDDR4 chip.
    """
    def __init__(self, clocks, log_level,
            auto_precharge=False, with_refresh=True, trace_reset=0, disable_delay=False,
            masked_write=True, double_rate_phy=False, finish_after_memtest=False, **kwargs):
        platform     = Platform(_io, clocks)
        sys_clk_freq = clocks["sys"]["freq_hz"]

        # SoCCore ----------------------------------------------------------------------------------
        super().__init__(platform,
            clk_freq      = sys_clk_freq,
            ident         = "LiteX Simulation",
            cpu_variant   = "lite",
            **kwargs)

        # CRG --------------------------------------------------------------------------------------
        self.submodules.crg = CRG(platform, clocks)

        # Debugging --------------------------------------------------------------------------------
        platform.add_debug(self, reset=trace_reset)

        # LPDDR4 -----------------------------------------------------------------------------------
        sdram_module = litedram_modules.MT53E256M16D1(sys_clk_freq, "1:8")
        pads = platform.request("lpddr4")
        sim_phy_cls = DoubleRateLPDDR4SimPHY if double_rate_phy else LPDDR4SimPHY
        self.submodules.ddrphy = sim_phy_cls(
            sys_clk_freq       = sys_clk_freq,
            aligned_reset_zero = True,
            masked_write       = masked_write,
        )

        for p in ["clk", "cke", "odt", "reset_n", "cs", "ca", "dq", "dqs", "dmi"]:
            self.comb += getattr(pads, p).eq(getattr(self.ddrphy.pads, p))

        controller_settings = ControllerSettings()
        controller_settings.auto_precharge = auto_precharge
        controller_settings.with_refresh = with_refresh

        self.add_sdram("sdram",
            phy                     = self.ddrphy,
            module                  = sdram_module,
            origin                  = self.mem_map["main_ram"],
            size                    = kwargs.get("max_sdram_size", 0x40000000),
            l2_cache_size           = kwargs.get("l2_size", 8192),
            l2_cache_min_data_width = kwargs.get("min_l2_data_width", 128),
            l2_cache_reverse        = False,
            controller_settings     = controller_settings
        )
        # Reduce memtest size for simulation speedup
        self.add_constant("MEMTEST_DATA_SIZE", 8*1024)
        self.add_constant("MEMTEST_ADDR_SIZE", 8*1024)

        # LPDDR4 Sim -------------------------------------------------------------------------------
        self.submodules.lpddr4sim = LPDDR4Sim(
            pads          = self.ddrphy.pads,
            cl            = self.sdram.controller.settings.phy.cl,
            cwl           = self.sdram.controller.settings.phy.cwl,
            sys_clk_freq  = sys_clk_freq,
            log_level     = log_level,
            disable_delay = disable_delay,
        )
        self.add_csr("lpddr4sim")

        self.add_constant("CONFIG_SIM_DISABLE_BIOS_PROMPT")
        if disable_delay:
            self.add_constant("CONFIG_DISABLE_DELAYS")
        if finish_after_memtest:
            self.submodules.ddrctrl = LiteDRAMCoreControl()
            self.add_csr("ddrctrl")
            self.sync += If(self.ddrctrl.init_done.storage, Finish())

        # Reuse DFITimingsChecker from phy/model.py
        nphases = self.sdram.controller.settings.phy.nphases
        timings = {"tCK": (1e9 / sys_clk_freq) / nphases}
        for name in _speedgrade_timings + _technology_timings:
            timings[name] = sdram_module.get(name)

        self.submodules.dfi_timings_checker = DFITimingsChecker(
            dfi          = self.ddrphy.dfi,
            nbanks       = 2**self.sdram.controller.settings.geom.bankbits,
            nphases      = nphases,
            timings      = timings,
            refresh_mode = sdram_module.timing_settings.fine_refresh_mode,
            memtype      = self.sdram.controller.settings.phy.memtype,
            verbose      = False,
        )

        # Debug info -------------------------------------------------------------------------------
        def dump(obj):
            print()
            print(" " + obj.__class__.__name__)
            print(" " + "-" * len(obj.__class__.__name__))
            d = obj if isinstance(obj, dict) else vars(obj)
            for var, val in d.items():
                if var == "self":
                    continue
                if isinstance(val, Signal):
                    val = "Signal(reset={})".format(val.reset.value)
                print("  {}: {}".format(var, val))

        print("=" * 80)
        dump(clocks)
        dump(self.ddrphy.settings)
        dump(sdram_module.geom_settings)
        dump(sdram_module.timing_settings)
        print()
        print("=" * 80)

# Build --------------------------------------------------------------------------------------------

def generate_gtkw_savefile(builder, vns, trace_fst):
    from litex.build.sim import gtkwave as gtkw

    dumpfile = os.path.join(builder.gateware_dir, "sim.{}".format("fst" if trace_fst else "vcd"))
    savefile = os.path.join(builder.gateware_dir, "sim.gtkw")
    soc = builder.soc
    wrphase = soc.sdram.controller.settings.phy.wrphase.reset.value

    with gtkw.GTKWSave(vns, savefile=savefile, dumpfile=dumpfile) as save:
        save.clocks()
        save.add(soc.bus.slaves["main_ram"], mappers=[gtkw.wishbone_sorter(), gtkw.wishbone_colorer()])
        save.fsm_states(soc)
        # all dfi signals
        save.add(soc.ddrphy.dfi, mappers=[gtkw.dfi_sorter(), gtkw.dfi_in_phase_colorer()])
        # each phase in separate group
        with save.gtkw.group("dfi phaseX", closed=True):
            for i, phase in enumerate(soc.ddrphy.dfi.phases):
                save.add(phase, group_name="dfi p{}".format(i), mappers=[
                    gtkw.dfi_sorter(phases=False),
                    gtkw.dfi_in_phase_colorer(),
                ])
        # only dfi command signals
        save.add(soc.ddrphy.dfi, group_name="dfi commands", mappers=[
            gtkw.regex_filter(gtkw.suffixes2re(["cas_n", "ras_n", "we_n"])),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        # only dfi data signals
        save.add(soc.ddrphy.dfi, group_name="dfi wrdata", mappers=[
            gtkw.regex_filter(["wrdata$", "p{}.*wrdata_en$".format(wrphase)]),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        save.add(soc.ddrphy.dfi, group_name="dfi wrdata_mask", mappers=[
            gtkw.regex_filter(gtkw.suffixes2re(["wrdata_mask"])),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        save.add(soc.ddrphy.dfi, group_name="dfi rddata", mappers=[
            gtkw.regex_filter(gtkw.suffixes2re(["rddata", "p0.*rddata_valid"])),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        # serialization
        with save.gtkw.group("serialization", closed=True):
            if isinstance(soc.ddrphy, DoubleRateLPDDR4SimPHY):
                ser_groups = [("out 1x", soc.ddrphy._out), ("out 2x", soc.ddrphy.out)]
            else:
                ser_groups = [("out", soc.ddrphy.out)]
            for name, out in ser_groups:
                save.group([out.cs, out.dqs_o[0], out.dqs_oe, out.dmi_o[0], out.dmi_oe],
                    group_name = name,
                    mappers = [
                        gtkw.regex_colorer({
                            "yellow": gtkw.suffixes2re(["cs"]),
                            "orange": ["_o[^e]"],
                            "red": gtkw.suffixes2re(["oe"]),
                        })
                    ]
                )
        with save.gtkw.group("deserialization", closed=True):
            if isinstance(soc.ddrphy, DoubleRateLPDDR4SimPHY):
                ser_groups = [("in 1x", soc.ddrphy._out), ("in 2x", soc.ddrphy.out)]
            else:
                ser_groups = [("in", soc.ddrphy.out)]
            for name, out in ser_groups:
                save.group([out.dq_i[0], out.dq_oe, out.dqs_i[0], out.dqs_oe],
                    group_name = name,
                    mappers = [gtkw.regex_colorer({
                        "yellow": ["dqs"],
                        "orange": ["dq[^s]"],
                    })]
                )
        # dram pads
        save.group([s for s in vars(soc.ddrphy.pads).values() if isinstance(s, Signal)],
            group_name = "pads",
            mappers = [
                gtkw.regex_filter(["_[io]$"], negate=True),
                gtkw.regex_sorter(gtkw.suffixes2re(["cke", "odt", "reset_n", "clk", "cs", "ca", "dq", "dqs", "dmi", "oe"])),
                gtkw.regex_colorer({
                    "yellow": gtkw.suffixes2re(["cs", "ca"]),
                    "orange": gtkw.suffixes2re(["dq", "dqs", "dmi"]),
                    "red": gtkw.suffixes2re(["oe"]),
                }),
            ],
        )

def main():
    parser = argparse.ArgumentParser(description="Generic LiteX SoC Simulation")
    builder_args(parser.add_argument_group(title="Builder"))
    soc_core_args(parser.add_argument_group(title="SoC Core"))
    group = parser.add_argument_group(title="LPDDR4 simulation")
    group.add_argument("--sdram-verbosity",      default=0,               help="Set SDRAM checker verbosity")
    group.add_argument("--trace",                action="store_true",     help="Enable Tracing")
    group.add_argument("--trace-fst",            action="store_true",     help="Enable FST tracing (default=VCD)")
    group.add_argument("--trace-start",          default=0,               help="Cycle to start tracing")
    group.add_argument("--trace-end",            default=-1,              help="Cycle to end tracing")
    group.add_argument("--trace-reset",          default=0,               help="Initial traceing state")
    group.add_argument("--sys-clk-freq",         default="50e6",          help="Core clock frequency")
    group.add_argument("--auto-precharge",       action="store_true",     help="Use DRAM auto precharge")
    group.add_argument("--no-refresh",           action="store_true",     help="Disable DRAM refresher")
    group.add_argument("--log-level",            default="all=INFO",      help="Set simulation logging level")
    group.add_argument("--disable-delay",        action="store_true",     help="Disable CPU delays")
    group.add_argument("--gtkw-savefile",        action="store_true",     help="Generate GTKWave savefile")
    group.add_argument("--no-masked-write",      action="store_true",     help="Use LPDDR4 WRITE instead of MASKED-WRITE")
    group.add_argument("--no-run",               action="store_true",     help="Don't run the simulation, just generate files")
    group.add_argument("--double-rate-phy",      action="store_true",     help="Use sim PHY with 2-stage serialization")
    group.add_argument("--finish-after-memtest", action="store_true",     help="Stop simulation after DRAM memory test")
    args = parser.parse_args()

    soc_kwargs     = soc_core_argdict(args)
    builder_kwargs = builder_argdict(args)

    sim_config = SimConfig()
    sys_clk_freq = int(float(args.sys_clk_freq))
    clocks = get_clocks(sys_clk_freq)
    clocks.add_clockers(sim_config)

    # Configuration --------------------------------------------------------------------------------
    if soc_kwargs["uart_name"] == "serial":
        soc_kwargs["uart_name"] = "sim"
        sim_config.add_module("serial2console", "serial")
    args.with_sdram = True
    soc_kwargs["integrated_main_ram_size"] = 0x0
    soc_kwargs["sdram_verbosity"]          = int(args.sdram_verbosity)

    # SoC ------------------------------------------------------------------------------------------
    soc = SimSoC(
        clocks          = clocks,
        auto_precharge  = args.auto_precharge,
        with_refresh    = not args.no_refresh,
        trace_reset     = int(args.trace_reset),
        log_level       = args.log_level,
        disable_delay   = args.disable_delay,
        masked_write    = not args.no_masked_write,
        double_rate_phy = args.double_rate_phy,
        finish_after_memtest = args.finish_after_memtest,
        **soc_kwargs)

    # Build/Run ------------------------------------------------------------------------------------
    builder_kwargs["csr_csv"] = "csr.csv"
    builder = Builder(soc, **builder_kwargs)
    build_kwargs = dict(
        sim_config  = sim_config,
        trace       = args.trace,
        trace_fst   = args.trace_fst,
        trace_start = int(args.trace_start),
        trace_end   = int(args.trace_end)
    )
    vns = builder.build(run=False, **build_kwargs)

    if args.gtkw_savefile:
        generate_gtkw_savefile(builder, vns, trace_fst=args.trace_fst)

    if not args.no_run:
        builder.build(build=False, **build_kwargs)

if __name__ == "__main__":
    main()



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.interconnect.csr import *

from litedram.common import *
from litedram.phy.dfi import *
from litedram.phy.rpc.basephy import BasePHY, bitpattern

# Xilinx Artix7 RPC PHY ----------------------------------------------------------------------------

class A7RPCPHY(BasePHY):
    def __init__(self, iodelay_clk_freq, **kwargs):
        self._rdly_dq_rst = CSR()
        self._rdly_dq_inc = CSR()

        self._db_enabled = CSRStorage(reset=1)
        self._dqs_enabled = CSRStorage(reset=1)

        kwargs.update(dict(
            write_ser_latency = 1,  # OSERDESE2 8:1 DDR (4 full-rate clocks)
            read_des_latency  = 2,  # ISERDESE2 NETWORKING
            phytype           = self.__class__.__name__,
        ))

        super().__init__(**kwargs)

        self.settings.delays = 32
        self.settings.read_leveling = True

        self.iodelay_clk_freq = iodelay_clk_freq
        iodelay_tap_average = {
            200e6: 78e-12,
            300e6: 52e-12,
            400e6: 39e-12,  # Only valid for -3 and -2/2E speed grades
        }
        self.half_sys8x_taps = math.floor(self.tck/(4*iodelay_tap_average[iodelay_clk_freq]))

    def do_clock_serialization(self, clk_1ck_out, clk_p, clk_n):
        clk = Signal()
        self.oserdese2_ddr(din=clk_1ck_out, dout=clk, clk="sys4x_180")
        self.specials += Instance("OBUFDS",
            i_I  = clk,
            o_O  = clk_p,
            o_OB = clk_n,
        )

    def do_stb_serialization(self, stb_1ck_out, stb):
        stb_out        = Signal()
        stb_in         = Signal()
        stb_in_delayed = Signal()
        stb_t          = Signal()

        self.stb_1ck_in = stb_1ck_in = Signal.like(stb_1ck_out)

        self.oserdese2_ddr(din=stb_1ck_out, dout=stb_out,
                           tin=Constant(0), tout=stb_t,
                           clk="sys4x_90")

        # Read path
        self.idelaye2(
            din=stb_in, dout=stb_in_delayed,
            rst=self.get_rst(0, self._rdly_dq_rst.re),
            inc=self.get_inc(0, self._rdly_dq_inc.re),
        )
        self.iserdese2_ddr(din=stb_in_delayed, dout=stb_1ck_in, clk="sys4x_180")

        self.specials += Instance("IOBUF",
            i_I   = stb_out,
            o_O   = stb_in,
            i_T   = stb_t,
            io_IO = stb,
        )

    def do_db_serialization(self, db_1ck_out, db_1ck_in, db_oe, db):
        for i in range(self.databits):
            db_out        = Signal()
            db_t          = Signal()
            db_in         = Signal()
            db_in_delayed = Signal()

            # Write path
            self.oserdese2_ddr(
                din=db_1ck_out[i], dout=db_out,
                tin=~(db_oe & self._db_enabled.storage), tout=db_t,
                clk="sys4x_90",
            )

            # Read path
            self.idelaye2(
                din=db_in, dout=db_in_delayed,
                rst=self.get_rst(i//8, self._rdly_dq_rst.re),
                inc=self.get_inc(i//8, self._rdly_dq_inc.re),
            )
            self.iserdese2_ddr(din=db_in_delayed, dout=db_1ck_in[i], clk="sys4x_180")

            self.specials += Instance("IOBUF",
                i_I   = db_out,
                o_O   = db_in,
                i_T   = db_t,
                io_IO = db[i],
            )

    def do_dqs_serialization(self, dqs_1ck_out, dqs_1ck_in, dqs_oe, dqs_p, dqs_n):
        for i in range(len(dqs_p)):
            dqs_out  = Signal()
            dqs_in   = Signal()
            dqs_t    = Signal()
            dqs_in_delayed  = Signal()

            self.oserdese2_ddr(
                clk="sys4x_180",
                din=dqs_1ck_out, dout=dqs_out,
                tin=~(dqs_oe & self._dqs_enabled.storage), tout=dqs_t,
            )

            # TODO: proper deserialization
            if i == 0:
                self.idelaye2(
                    din=dqs_in, dout=dqs_in_delayed,
                    rst=self.get_rst(i//8, self._rdly_dq_rst.re),
                    inc=self.get_inc(i//8, self._rdly_dq_inc.re),
                )
                self.iserdese2_ddr(
                    clk="sys4x_90",
                    din=dqs_in_delayed, dout=dqs_1ck_in)

            self.specials += Instance("IOBUFDS",
                i_T    = dqs_t,
                i_I    = dqs_out,
                o_O    = dqs_in,
                io_IO  = dqs_p[i],
                io_IOB = dqs_n[i],
            )

    def do_cs_serialization(self, cs_n_1ck_out, cs_n):
        self.oserdese2_ddr(din=cs_n_1ck_out, dout=cs_n, clk="sys4x_90")

    def idelaye2(self, *, din, dout, init=0, rst=None, inc=None):
        assert not ((rst is None) ^ (inc is None))
        fixed = rst is None

        params = dict(
            p_SIGNAL_PATTERN        = "DATA",
            p_DELAY_SRC             = "IDATAIN",
            p_CINVCTRL_SEL          = "FALSE",
            p_HIGH_PERFORMANCE_MODE = "TRUE",
            p_REFCLK_FREQUENCY      = self.iodelay_clk_freq/1e6,
            p_PIPE_SEL              = "FALSE",
            p_IDELAY_VALUE          = init,
            p_IDELAY_TYPE           = "FIXED",
            i_IDATAIN  = din,
            o_DATAOUT  = dout,
        )

        if not fixed:
            params.update(dict(
                p_IDELAY_TYPE  = "VARIABLE",
                i_C        = ClockSignal(),
                i_LD       = rst,
                i_CE       = inc,
                i_LDPIPEEN = 0,
                i_INC      = 1,
            ))

        self.specials += Instance("IDELAYE2", **params)

    def odelaye2(self, *, din, dout, init=0, rst=None, inc=None):  # Not available for Artix7
        assert not ((rst is None) ^ (inc is None))
        fixed = rst is not None

        params = dict(
            p_SIGNAL_PATTERN        = "DATA",
            p_DELAY_SRC             = "ODATAIN",
            p_CINVCTRL_SEL          = "FALSE",
            p_HIGH_PERFORMANCE_MODE = "TRUE",
            p_REFCLK_FREQUENCY      = self.iodelay_clk_freq/1e6,
            p_PIPE_SEL              = "FALSE",
            p_ODELAY_VALUE          = init,
            p_ODELAY_TYPE           = "FIXED",
            i_ODATAIN  = din,
            o_DATAOUT  = dout,
        )

        if not fixed:
            params.update(dict(
                p_ODELAY_TYPE  = "VARIABLE",
                i_C        = ClockSignal(),
                i_LD       = rst,
                i_CE       = inc,
                i_LDPIPEEN = 0,
                i_INC      = 1,
            ))

        self.specials += Instance("ODELAYE2", **params)

    def oserdese2_ddr(self, *, din, dout, clk, tin=None, tout=None):
        assert self.nphases == 4
        assert not ((tin is None) ^ (tout is None))

        params = dict(
            p_SERDES_MODE    = "MASTER",
            p_DATA_WIDTH     = 2*self.nphases,
            p_TRISTATE_WIDTH = 1,
            p_DATA_RATE_OQ   = "DDR",
            p_DATA_RATE_TQ   = "BUF",
            i_RST    = ResetSignal(),
            i_CLK    = ClockSignal(clk),
            i_CLKDIV = ClockSignal("sys"),
            o_OQ     = dout,
            i_OCE    = 1,
        )

        for i in range(2*self.nphases):
            params["i_D{}".format(i+1)] = din[i]

        if tin is not None:
            # with DATA_RATE_TQ=BUF tristate is asynchronous, so we need to delay it
            tin_d = Signal()
            self.sync += tin_d.eq(tin)

            # register it on the CLKDIV (as it would be too short for 180 deg shifted clk)
            tin_cdc = Signal()
            sd_clkdiv = getattr(self.sync, clk)
            sd_clkdiv += tin_cdc.eq(tin_d)

            params.update(dict(i_TCE=1, i_T1=tin_cdc, o_TQ=tout))

        self.specials += Instance("OSERDESE2", **params)

    def iserdese2_ddr(self, *, din, dout, clk):
        assert self.nphases == 4

        params = dict(
            p_SERDES_MODE    = "MASTER",
            p_INTERFACE_TYPE = "NETWORKING",  # TODO: try using MEMORY mode?
            p_DATA_WIDTH     = 2*self.nphases,
            p_DATA_RATE      = "DDR",
            p_NUM_CE         = 1,
            p_IOBDELAY       = "IFD",
            i_RST     = ResetSignal(),
            i_CLK     = ClockSignal(clk),
            i_CLKB    = ~ClockSignal(clk),
            i_CLKDIV  = ClockSignal("sys"),
            i_BITSLIP = 0,
            i_CE1     = 1,
            i_DDLY    = din,
        )

        for i in range(2*self.nphases):
            # invert order
            params["o_Q{}".format(i+1)] = dout[(2*self.nphases - 1) - i]

        self.specials += Instance("ISERDESE2", **params)



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2016-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
# SPDX-License-Identifier: BSD-2-Clause

"""LiteDRAM Refresher."""

from migen import *
from migen.genlib.misc import timeline

from litex.soc.interconnect import stream

from litedram.core.multiplexer import *

# RefreshExecuter ----------------------------------------------------------------------------------

class RefreshExecuter(Module):
    """Refresh Executer

    Execute the refresh sequence to the DRAM:
    - Send a "Precharge All" command
    - Wait tRP
    - Send an "Auto Refresh" command
    - Wait tRFC
    """
    def __init__(self, cmd, trp, trfc):
        self.start = Signal()
        self.done  = Signal()

        # # #

        self.sync += [
            cmd.a.eq(  0),
            cmd.ba.eq( 0),
            cmd.cas.eq(0),
            cmd.ras.eq(0),
            cmd.we.eq( 0),
            self.done.eq(0),
            # Wait start
            timeline(self.start, [
                # Precharge All
                (0, [
                    cmd.a.eq(  2**10),
                    cmd.ba.eq( 0),
                    cmd.cas.eq(0),
                    cmd.ras.eq(1),
                    cmd.we.eq( 1)
                ]),
                # Auto Refresh after tRP
                (trp, [
                    cmd.a.eq(  2**10),  # all banks in LPDDR4/DDR5, ignored in other memories
                    cmd.ba.eq( 0),
                    cmd.cas.eq(1),
                    cmd.ras.eq(1),
                    cmd.we.eq( 0),
                ]),
                # Done after tRP + tRFC
                (trp + trfc, [
                    cmd.a.eq(  0),
                    cmd.ba.eq( 0),
                    cmd.cas.eq(0),
                    cmd.ras.eq(0),
                    cmd.we.eq( 0),
                    self.done.eq(1),
                ]),
            ])
        ]

# RefreshSequencer ---------------------------------------------------------------------------------

class RefreshSequencer(Module):
    """Refresh Sequencer

    Sequence N refreshs to the DRAM.
    """
    def __init__(self, cmd, trp, trfc, postponing=1):
        self.start = Signal()
        self.done  = Signal()

        # # #

        executer = RefreshExecuter(cmd, trp, trfc)
        self.submodules += executer

        count = Signal(bits_for(postponing), reset=postponing-1)
        self.sync += [
            If(self.start,
                count.eq(count.reset)
            ).Elif(executer.done,
                If(count != 0,
                    count.eq(count - 1)
                )
            )
        ]
        self.comb += executer.start.eq(self.start | (count != 0))
        self.comb += self.done.eq(executer.done & (count == 0))

# RefreshTimer -------------------------------------------------------------------------------------

class RefreshTimer(Module):
    """Refresh Timer

    Generate periodic pulses (tREFI period) to trigger DRAM refresh.
    """
    def __init__(self, trefi):
        self.wait  = Signal()
        self.done  = Signal()
        self.count = Signal(bits_for(trefi))

        # # #

        done  = Signal()
        count = Signal(bits_for(trefi), reset=trefi-1)

        self.sync += [
            If(self.wait & ~self.done,
                count.eq(count - 1)
            ).Else(
                count.eq(count.reset)
            )
        ]
        self.comb += [
            done.eq(count == 0),
            self.done.eq(done),
            self.count.eq(count)
        ]

# RefreshPostponer -------------------------------------------------------------------------------

class RefreshPostponer(Module):
    """Refresh Postponer

    Postpone N Refresh requests and generate a request when N is reached.
    """
    def __init__(self, postponing=1):
        self.req_i = Signal()
        self.req_o = Signal()

        # # #

        count = Signal(bits_for(postponing), reset=postponing-1)
        self.sync += [
            self.req_o.eq(0),
            If(self.req_i,
                count.eq(count - 1),
                If(count == 0,
                    count.eq(count.reset),
                    self.req_o.eq(1)
                )
            )
        ]

# ZQCSExecuter ----------------------------------------------------------------------------------

class ZQCSExecuter(Module):
    """ZQ Short Calibration Executer

    Execute the ZQCS sequence to the DRAM:
    - Send a "Precharge All" command
    - Wait tRP
    - Send an "ZQ Short Calibration" command
    - Wait tZQCS
    """
    def __init__(self, cmd, trp, tzqcs):
        self.start = Signal()
        self.done  = Signal()

        # # #

        self.sync += [
            # Note: Don't set cmd to 0 since already done in RefreshExecuter
            self.done.eq(0),
            # Wait start
            timeline(self.start, [
                # Precharge All
                (0, [
                    cmd.a.eq(  2**10),
                    cmd.ba.eq( 0),
                    cmd.cas.eq(0),
                    cmd.ras.eq(1),
                    cmd.we.eq( 1)
                ]),
                # ZQ Short Calibration after tRP
                (trp, [
                    cmd.a.eq(  0),
                    cmd.ba.eq( 0),
                    cmd.cas.eq(0),
                    cmd.ras.eq(0),
                    cmd.we.eq( 1),
                ]),
                # Done after tRP + tZQCS
                (trp + tzqcs, [
                    cmd.a.eq(  0),
                    cmd.ba.eq( 0),
                    cmd.cas.eq(0),
                    cmd.ras.eq(0),
                    cmd.we.eq( 0),
                    self.done.eq(1)
                ]),
            ])
        ]

# Refresher ----------------------------------------------------------------------------------------

class Refresher(Module):
    """Refresher

    Manage DRAM refresh.

    The DRAM needs to be periodically refreshed with a tREFI period to avoid data corruption. During
    a refresh, the controller send a "Precharge All" command to close and precharge all rows and then
    send a "Auto Refresh" command.

    Before executing the refresh, the Refresher advertises the Controller that a refresh should occur,
    this allows the Controller to finish the current transaction and block next transactions. Once all
    transactions are done, the Refresher can execute the refresh Sequence and release the Controller.

    """
    def __init__(self, settings, clk_freq, zqcs_freq=1e0, postponing=1):
        assert postponing <= 8
        abits  = settings.geom.addressbits
        babits = settings.geom.bankbits + log2_int(settings.phy.nranks)
        self.cmd = cmd = stream.Endpoint(cmd_request_rw_layout(a=abits, ba=babits))

        # # #

        wants_refresh = Signal()
        wants_zqcs    = Signal()

        # Refresh Timer ----------------------------------------------------------------------------
        if settings.timing.tREFI < 100: # FIXME: Reduce Margin.
            raise ValueError("Clk/tREFI is ratio too low , please increase Clk frequency or disable Refresh.")
        timer = RefreshTimer(settings.timing.tREFI)
        self.submodules.timer = timer
        self.comb += timer.wait.eq(~timer.done)

        # Refresh Postponer ------------------------------------------------------------------------
        postponer = RefreshPostponer(postponing)
        self.submodules.postponer = postponer
        self.comb += postponer.req_i.eq(self.timer.done)
        self.comb += wants_refresh.eq(postponer.req_o)

        # Refresh Sequencer ------------------------------------------------------------------------
        sequencer = RefreshSequencer(cmd, settings.timing.tRP, settings.timing.tRFC, postponing)
        self.submodules.sequencer = sequencer

        if settings.timing.tZQCS is not None:
            # ZQCS Timer ---------------------------------------------------------------------------
            zqcs_timer = RefreshTimer(int(clk_freq/zqcs_freq))
            self.submodules.zqcs_timer = zqcs_timer
            self.comb += wants_zqcs.eq(zqcs_timer.done)

            # ZQCS Executer ------------------------------------------------------------------------
            zqcs_executer = ZQCSExecuter(cmd, settings.timing.tRP, settings.timing.tZQCS)
            self.submodules.zqs_executer = zqcs_executer
            self.comb += zqcs_timer.wait.eq(~zqcs_executer.done)

        # Refresh FSM ------------------------------------------------------------------------------
        self.submodules.fsm = fsm = FSM()
        fsm.act("IDLE",
            If(settings.with_refresh,
                If(wants_refresh,
                    NextState("WAIT-BANK-MACHINES")
                )
            )
        )
        fsm.act("WAIT-BANK-MACHINES",
            cmd.valid.eq(1),
            If(cmd.ready,
                sequencer.start.eq(1),
                NextState("DO-REFRESH")
            )
        )
        if settings.timing.tZQCS is None:
            fsm.act("DO-REFRESH",
                cmd.valid.eq(1),
                If(sequencer.done,
                    cmd.valid.eq(0),
                    cmd.last.eq(1),
                    NextState("IDLE")
                )
            )
        else:
            fsm.act("DO-REFRESH",
                cmd.valid.eq(1),
                If(sequencer.done,
                    If(wants_zqcs,
                        zqcs_executer.start.eq(1),
                        NextState("DO-ZQCS")
                    ).Else(
                        cmd.valid.eq(0),
                        cmd.last.eq(1),
                        NextState("IDLE")
                    )
                )
            )
            fsm.act("DO-ZQCS",
                cmd.valid.eq(1),
                If(zqcs_executer.done,
                    cmd.valid.eq(0),
                    cmd.last.eq(1),
                    NextState("IDLE")
                )
            )



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2016-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2018 John Sully <john@csquare.ca>
# Copyright (c) 2018 bunnie <bunnie@kosagi.com>
# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import math
from functools import reduce
from operator import add
from collections import OrderedDict
from typing import Union, Optional

from migen import *

from litex.soc.interconnect import stream

# Helpers ------------------------------------------------------------------------------------------

burst_lengths = {
    "SDR":    1,
    "DDR":    4,
    "LPDDR":  4,
    "DDR2":   4,
    "DDR3":   8,
    "RPC":    16,
    "DDR4":   8,
    "LPDDR4": 16,
    "LPDDR5": 16,
}

def get_default_cl_cwl(memtype, tck):
    f_to_cl_cwl = OrderedDict()
    if memtype == "SDR":
        f_to_cl_cwl[100e6] = (2, None)
        f_to_cl_cwl[133e6] = (3, None)
    elif memtype == "DDR2":
        f_to_cl_cwl[400e6]  = (3, 2)
        f_to_cl_cwl[533e6]  = (4, 3)
        f_to_cl_cwl[677e6]  = (5, 4)
        f_to_cl_cwl[800e6]  = (6, 5)
        f_to_cl_cwl[1066e6] = (7, 5)
    elif memtype == "DDR3":
        f_to_cl_cwl[800e6]  = ( 6, 5)
        f_to_cl_cwl[1066e6] = ( 7, 6)
        f_to_cl_cwl[1333e6] = (10, 7)
        f_to_cl_cwl[1600e6] = (11, 8)
        f_to_cl_cwl[1866e6] = (13, 9)
    elif memtype == "DDR4":
        f_to_cl_cwl[1333e6] = (9,   9)
        f_to_cl_cwl[1600e6] = (11,  9)
        f_to_cl_cwl[1866e6] = (13, 10)
        f_to_cl_cwl[2133e6] = (15, 11)
        f_to_cl_cwl[2400e6] = (16, 12)
        f_to_cl_cwl[2666e6] = (18, 14)
    else:
        raise ValueError
    for f, (cl, cwl) in f_to_cl_cwl.items():
        m = 2 if "DDR" in memtype else 1
        if tck >= m/f:
            return cl, cwl
    raise ValueError

def get_default_cl(memtype, tck):
    cl, _ = get_default_cl_cwl(memtype, tck)
    return cl

def get_default_cwl(memtype, tck):
    _, cwl = get_default_cl_cwl(memtype, tck)
    return cwl

def get_sys_latency(nphases, cas_latency):
    return math.ceil(cas_latency/nphases)

def get_sys_phase(nphases, sys_latency, cas_latency):
    return sys_latency*nphases - cas_latency

# PHY Pads Transformers ----------------------------------------------------------------------------

class PHYPadsReducer:
    """PHY Pads Reducer

    Reduce DRAM pads to only use specific modules.

    For testing purposes, we often need to use only some of the DRAM modules. PHYPadsReducer allows
    selecting specific modules and avoid re-definining dram pins in the Platform for this.
    """
    def __init__(self, pads, modules, with_cat=False):
        self.pads     = pads
        self.modules  = modules
        self.with_cat = with_cat

    def __getattr__(self, name):
        if name in ["dq"]:
            r = Array([getattr(self.pads, name)[8*i + j]
                for i in self.modules
                for j in range(8)])
            return r if not self.with_cat else Cat(r)
        if name in ["dm", "dqs", "dqs_p", "dqs_n"]:
            r = Array([getattr(self.pads, name)[i] for i in self.modules])
            return r if not self.with_cat else Cat(r)
        else:
            return getattr(self.pads, name)

class PHYPadsCombiner:
    """PHY Pads Combiner

    Combine DRAM pads from fully dissociated chips in a unique DRAM pads structure.

    Most generally, DRAM chips are sharing command/address lines between chips (using a fly-by
    topology since DDR3). On some boards, the DRAM chips are using separate command/address lines
    and this combiner can be used to re-create a single pads structure (that will be compatible with
    LiteDRAM's PHYs) to create a single DRAM controller from multiple fully dissociated DRAMs chips.
    """
    def __init__(self, pads):
        if not isinstance(pads, list):
            self.groups = [pads]
        else:
            self.groups = pads
        self.sel = 0

    def sel_group(self, n):
        self.sel = n

    def __getattr__(self, name):
        if name in ["dm", "dq", "dqs", "dqs_p", "dqs_n"]:
            return Array([getattr(self.groups[j], name)[i]
                for i in range(len(getattr(self.groups[0], name)))
                for j in range(len(self.groups))])
        else:
            return getattr(self.groups[self.sel], name)

# BitSlip ------------------------------------------------------------------------------------------

class BitSlip(Module):
    def __init__(self, dw, i=None, o=None, rst=None, slp=None, cycles=1):
        self.i   = Signal(dw) if i is None else i
        self.o   = Signal(dw) if o is None else o
        self.rst = Signal()   if rst is None else rst
        self.slp = Signal()   if slp is None else slp
        assert cycles >= 1

        # # #

        value = Signal(max=cycles*dw, reset=cycles*dw-1)
        self.sync += If(self.slp, value.eq(value + 1))
        self.sync += If(self.rst, value.eq(value.reset))

        r = Signal((cycles+1)*dw, reset_less=True)
        self.sync += r.eq(Cat(r[dw:], self.i))
        cases = {}
        for i in range(cycles*dw):
            cases[i] = self.o.eq(r[i+1:dw+i+1])
        self.comb += Case(value, cases)

# TappedDelayLine ----------------------------------------------------------------------------------

class TappedDelayLine(Module):
    def __init__(self, signal=None, ntaps=1):
        self.input = Signal() if signal is None else signal
        self.taps  = Array(Signal.like(self.input) for i in range(ntaps))
        for i in range(ntaps):
            self.sync += self.taps[i].eq(self.input if i == 0 else self.taps[i-1])
        self.output = self.taps[-1]

# DQS Pattern --------------------------------------------------------------------------------------

class DQSPattern(Module):
    def __init__(self, preamble=None, postamble=None, wlevel_en=0, wlevel_strobe=0, register=False):
        self.preamble  = Signal() if preamble  is None else preamble
        self.postamble = Signal() if postamble is None else postamble
        self.o = Signal(8)

        # # #

        # DQS Pattern transmitted as LSB-first.

        self.comb += [
            self.o.eq(0b01010101),
            If(self.preamble,
                self.o.eq(0b00010101)
            ),
            If(self.postamble,
                self.o.eq(0b01010100)
            ),
            If(wlevel_en,
                self.o.eq(0b00000000),
                If(wlevel_strobe,
                    self.o.eq(0b00000001)
                )
            )
        ]
        if register:
            o = Signal.like(self.o)
            self.sync += o.eq(self.o)
            self.o = o

# Settings -----------------------------------------------------------------------------------------

class Settings:
    def set_attributes(self, attributes):
        for k, v in attributes.items():
            setattr(self, k, v)


class PhySettings(Settings):
    def __init__(self,
            phytype: str,
            memtype: str,  # SDR, DDR, DDR2, ...
            databits: int,  # number of DQ lines
            dfi_databits: int,  # per-phase DFI data width
            nphases: int,  # number of DFI phases
            rdphase: Union[int, Signal],  # phase on which READ command will be issued by MC
            wrphase: Union[int, Signal],  # phase on which WRITE command will be issued by MC
            cl: int,  # latency (DRAM clk) from READ command to first data
            read_latency: int,  # latency (MC clk) from DFI.rddata_en to DFI.rddata_valid
            write_latency: int,  # latency (MC clk) from DFI.wrdata_en to DFI.wrdata
            strobes: Optional[int] = None,  # number of DQS lines
            nranks: int = 1,  # number of DRAM ranks
            cwl: Optional[int] = None,  # latency (DRAM clk) from WRITE command to first data
            cmd_latency: Optional[int] = None,  # additional command latency (MC clk)
            cmd_delay: Optional[int] = None,  # used to force cmd delay during initialization in BIOS
            bitslips: int = 0,  # number of write/read bitslip taps
            delays: int = 0,  # number of write/read delay taps
            # PHY training capabilities
            write_leveling: bool = False,
            write_dq_dqs_training: bool = False,
            write_latency_calibration: bool = False,
            read_leveling: bool = False,
        ):
        if strobes is None:
            strobes = databits // 8
        self.set_attributes(locals())
        self.cwl = cl if cwl is None else cwl
        self.is_rdimm = False

    # Optional DDR3/DDR4 electrical settings:
    # rtt_nom: Non-Writes on-die termination impedance
    # rtt_wr: Writes on-die termination impedance
    # ron: Output driver impedance
    # tdqs: Termination Data Strobe enable.
    def add_electrical_settings(self, rtt_nom=None, rtt_wr=None, ron=None, tdqs=None):
        assert self.memtype in ["DDR3", "DDR4"]
        if rtt_nom is not None:
            self.rtt = rtt_nom
        if rtt_wr is not None:
            self.rtt_wr = rtt_wr
        if ron is not None:
            self.ron = ron
        if tdqs is not None:
            self.tdqs = tdqs

    # Optional RDIMM configuration
    def set_rdimm(self, tck, rcd_pll_bypass, rcd_ca_cs_drive, rcd_odt_cke_drive, rcd_clk_drive):
        assert self.memtype == "DDR4"
        self.is_rdimm = True
        self.set_attributes(locals())

class GeomSettings(Settings):
    def __init__(self, bankbits, rowbits, colbits):
        self.set_attributes(locals())
        self.addressbits = max(rowbits, colbits)


class TimingSettings(Settings):
    def __init__(self, tRP, tRCD, tWR, tWTR, tREFI, tRFC, tFAW, tCCD, tRRD, tRC, tRAS, tZQCS):
        self.set_attributes(locals())

# Layouts/Interface --------------------------------------------------------------------------------

def cmd_layout(address_width):
    return [
        ("valid",            1, DIR_M_TO_S),
        ("ready",            1, DIR_S_TO_M),
        ("we",               1, DIR_M_TO_S),
        ("addr", address_width, DIR_M_TO_S),
        ("lock",             1, DIR_S_TO_M), # only used internally

        ("wdata_ready",      1, DIR_S_TO_M),
        ("rdata_valid",      1, DIR_S_TO_M)
    ]

def data_layout(data_width):
    return [
        ("wdata",       data_width, DIR_M_TO_S),
        ("wdata_we", data_width//8, DIR_M_TO_S),
        ("rdata",       data_width, DIR_S_TO_M)
    ]

def cmd_request_layout(a, ba):
    return [
        ("a",     a),
        ("ba",   ba),
        ("cas",   1),
        ("ras",   1),
        ("we",    1)
    ]

def cmd_request_rw_layout(a, ba):
    return cmd_request_layout(a, ba) + [
        ("is_cmd", 1),
        ("is_read", 1),
        ("is_write", 1)
    ]


class LiteDRAMInterface(Record):
    def __init__(self, address_align, settings):
        rankbits = log2_int(settings.phy.nranks)
        self.address_align = address_align
        self.address_width = settings.geom.rowbits + settings.geom.colbits + rankbits - address_align
        self.data_width    = settings.phy.dfi_databits*settings.phy.nphases
        self.nbanks   = settings.phy.nranks*(2**settings.geom.bankbits)
        self.nranks   = settings.phy.nranks
        self.settings = settings

        layout = [("bank"+str(i), cmd_layout(self.address_width)) for i in range(self.nbanks)]
        layout += data_layout(self.data_width)
        Record.__init__(self, layout)

# Ports --------------------------------------------------------------------------------------------

def cmd_description(address_width):
    return [
        ("we",               1), # Write (1) or Read (0).
        ("addr", address_width)  # Address (in Controller's words).
    ]

def wdata_description(data_width):
    return [
        ("data",    data_width), # Write Data.
        ("we",   data_width//8), # Write Data byte enable.
    ]

def rdata_description(data_width):
    return [("data", data_width)] # Read Data.

class LiteDRAMNativePort(Settings):
    def __init__(self, mode, address_width, data_width, clock_domain="sys", id=0):
        self.set_attributes(locals())

        self.flush = Signal()
        self.lock  = Signal()

        self.cmd   = stream.Endpoint(cmd_description(address_width))
        self.wdata = stream.Endpoint(wdata_description(data_width))
        self.rdata = stream.Endpoint(rdata_description(data_width))

        # retro-compatibility # FIXME: remove
        self.aw = self.address_width
        self.dw = self.data_width
        self.cd = self.clock_domain

    def get_bank_address(self, bank_bits, cba_shift):
        cba_upper = cba_shift + bank_bits
        return self.cmd.addr[cba_shift:cba_upper]

    def get_row_column_address(self, bank_bits, rca_bits, cba_shift):
        cba_upper = cba_shift + bank_bits
        if cba_shift < rca_bits:
            if cba_shift:
                return Cat(self.cmd.addr[:cba_shift], self.cmd.addr[cba_upper:])
            else:
                return self.cmd.addr[cba_upper:]
        else:
            return self.cmd.addr[:cba_shift]

    def connect(self, port):
        return [
            self.cmd.connect(port.cmd),
            self.wdata.connect(port.wdata),
            port.rdata.connect(self.rdata),
            port.flush.eq(self.flush),
            self.lock.eq(port.lock),
        ]

class LiteDRAMNativeWritePort(LiteDRAMNativePort):
    def __init__(self, *args, **kwargs):
        LiteDRAMNativePort.__init__(self, "write", *args, **kwargs)


class LiteDRAMNativeReadPort(LiteDRAMNativePort):
    def __init__(self, *args, **kwargs):
        LiteDRAMNativePort.__init__(self, "read", *args, **kwargs)


# Timing Controllers -------------------------------------------------------------------------------

class tXXDController(Module):
    def __init__(self, txxd):
        self.valid = valid = Signal()
        self.ready = ready = Signal(reset=txxd is None)
        ready.attr.add("no_retiming")

        # # #

        if txxd is not None:
            count = Signal(max=max(txxd, 2))
            self.sync += \
                If(valid,
                    count.eq(txxd - 1),
                    If((txxd - 1) == 0,
                        ready.eq(1)
                    ).Else(
                        ready.eq(0)
                    )
                ).Elif(~ready,
                    count.eq(count - 1),
                    If(count == 1,
                        ready.eq(1)
                    )
                )


class tFAWController(Module):
    def __init__(self, tfaw):
        self.valid = valid = Signal()
        self.ready = ready = Signal(reset=1)
        ready.attr.add("no_retiming")

        # # #

        if tfaw is not None:
            count  = Signal(max=max(tfaw, 2))
            window = Signal(tfaw)
            self.sync += window.eq(Cat(valid, window))
            self.comb += count.eq(reduce(add, [window[i] for i in range(tfaw)]))
            self.sync += \
                If(count < 4,
                    If(count == 3,
                        ready.eq(~valid)
                    ).Else(
                        ready.eq(1)
                    )
                )



# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import re
import enum

from migen import *

@enum.unique
class BankOrganization(enum.IntEnum):
    """Internal organization (architecture) of banks as set in MR[3]"""
    BG = 0b00   # 4 banks, 4 bank groups   (>3200 Mbps)
    B8 = 0b01   # 8 banks, no bank groups  (all data rates, BL32 only)
    B16 = 0b10  # 16 banks, no bank groups (<=3200 Mbps)

@enum.unique
class SpecialCmd(enum.IntEnum):
    """Codes for special commands encoded in DFI ZQC command

    The number of possible commands in LPDDR5 is too big to encode them
    in DFI in the regular way. DFI ZQC command is used to encode several
    special commands depending on the value of DFI.bank using DFI.address
    for additional data.

    NOP can be used to actually send NOP to the DRAM. By default we send
    DESELECT because LiteDRAM keeps holding CS_n low all the time, so we
    would be sending NOPs (toggling LPDDR5 CS pad) all the time.
    """
    MPC = 0
    MRR = 1
    NOP = 2

@enum.unique
class MPC(enum.IntEnum):
    """Op codes for LPDDR5 multipurpose command

    DFI ZQC command is used to send LPDDR5 MPC. DFI address A[7:0] is translated
    to MPC operand OP[7:0]. DFI bank address BA should be equal to SpecialCmd.MPC.
    """
    WCK2DQI_START = 0b10000001
    WCK2DQI_STOP  = 0b10000010
    WCK2DQO_START = 0b10000011
    WCK2DQO_STOP  = 0b10000100
    ZQC_START     = 0b10000101
    ZQC_LATCH     = 0b10000110
    # all others reserved


def dfi_cmd(dfi_phase):
    return Cat(~dfi_phase.we_n, ~dfi_phase.ras_n, ~dfi_phase.cas_n)

CMD = {  # cas, ras, we (2, 1, 0)
    "NOP": 0b000,
    "ACT": 0b010,
    "RD":  0b100,
    "WR":  0b101,
    "PRE": 0b011,
    "REF": 0b110,
    "ZQC": 0b001,
    "MRS": 0b111,
}


@enum.unique
class WCKSyncType(enum.IntEnum):
    """Corresponds to CAS WCK sync flags"""
    WR = 1
    RD = 2
    FS = 3


class DFIPhaseAdapter(Module):
    """Translates DFI phase into LPDDR5 command (2 or 4 CK edges)

    In LPDDR5 a "full command" may consist of 1 or 2 commands. Each command then consists
    of values for 2 consecutive clock edges. For DFI commands that require only 1 LPDDR5
    command, a DESELECT is inserted in on the first 2 CK edges to simplify timing calculations.

    Parameters
    ----------
    dfi_phase : Record(dfi.phase_description), in
        Input from a single DFI phase.
    masked_write : bool or Signal(1)
        Specifies how DFI write command (cas_n=0, ras_n=1, we_n=0) is interpreted, either
        as LPDDR5 WRITE16 or MASKE-WRITE.

    Attributes
    ----------
    cs : Signal(2), out
        Values of CS on 2 subsequent DRAM SDR CK cycles.
    ca : Array(4, Signal(6)), out
        Values of CA[6:0] on 4 subsequent DRAM DDR clock edges.
    valid : Signal, out
        Indicates that a valid command is presented on the `cs` and `ca` outputs.
    wck_sync_done: Signal, in
        Indicates whether WCK synchronization has already been done. PHY must drive
        this signal to control if CAS commands are sent with WCK sync bits.
    wck_sync: Signal(2), out
        Indicates that a CAS command with WCK sync is being sent in this cycle.
        PHY must use this signal to update `wck_sync_done`. The value is one of
        `WCKSyncType` enum, 0 means no WCK sync.
    """
    def __init__(self, dfi_phase, masked_write=True):
        assert isinstance(masked_write, (bool, Signal)), "Use boolean (static) or Signal (dynamic)"
        if isinstance(masked_write, bool):
            masked_write = int(masked_write)
        else:
            assert len(masked_write) == 1

        self.cs = Signal(2)
        self.ca = Array([Signal(7) for _ in range(4)])
        self.valid = Signal()
        self.wck_sync_done = Signal()
        self.wck_sync = Signal(max=len(WCKSyncType))

        # # #

        self.submodules.cmd1 = Command(dfi_phase, self.wck_sync)
        self.submodules.cmd2 = Command(dfi_phase, self.wck_sync)
        self.comb += [
            self.cs[0].eq(self.cmd1.cs),
            self.cs[1].eq(self.cmd2.cs),
            self.ca[0].eq(self.cmd1.ca[0]),
            self.ca[1].eq(self.cmd1.ca[1]),
            self.ca[2].eq(self.cmd2.ca[0]),
            self.ca[3].eq(self.cmd2.ca[1]),
        ]

        def wck_sync(type):
            return If(self.wck_sync_done == 0,
                self.wck_sync.eq(getattr(WCKSyncType, type.upper())),
            )

        def cmds(*cmd, valid=1):
            if len(cmd) == 1:
                ops = self.cmd1.set("DES") + self.cmd2.set(cmd[0])
            elif len(cmd) == 2:
                ops = self.cmd1.set(cmd[0]) + self.cmd2.set(cmd[1])
            else:
                raise ValueError(cmd)
            return ops + [self.valid.eq(valid)]

        # TODO: we don't actually need CAS command if WCK sync is on, but for now
        # send it to avoid problems when tracking module timings
        deselect = cmds("DES", "DES", valid=0)
        self.comb += If(dfi_phase.cs_n == 0,
            Case(dfi_cmd(dfi_phase), {
                CMD["ACT"]: cmds("ACT-1", "ACT-2"),
                CMD["RD"]: [*cmds("CAS", "RD16"), wck_sync("RD")],
                CMD["WR"]:  Case(masked_write, {
                    0: [*cmds("CAS", "WR16"), wck_sync("WR")],
                    1: [*cmds("CAS", "MWR"), wck_sync("WR")],
                }),
                CMD["PRE"]: cmds("PRE"),
                CMD["REF"]: cmds("REF"),
                CMD["ZQC"]: Case(dfi_phase.bank, {
                    SpecialCmd.MPC: cmds("MPC"),
                    SpecialCmd.MRR: [*cmds("CAS", "MRR"), wck_sync("RD")],
                    SpecialCmd.NOP: cmds("NOP"),
                    "default": deselect,
                }),
                CMD["MRS"]: cmds("MRW-1", "MRW-2"),
                "default": deselect,
            })
        )


class Command(Module):
    """LPDDR5 command decoder

    Decode commands from a DFI phase into LPDDR5 command consisting of 1 CS value
    and 2 CA[6:0] values. CA values are then to be sent over 2 CK edges (DDR),
    while CS is an SDR signal.

    Some LPDDR5 commands may consist of 2 separate "small commands", resulting in
    the command being actually sent over 2 CK cycles = 4 edges (e.g. ACT consists
    of ACTIVATE-1 and ACTIVATE-2).

    Attributes
    ----------
    dfi : Record(dfi.phase_description), in
        Input from single DFI phase.
    cs : Signal(), out
        CS value for that CK SDR cycle
    ca : Array(2, Signal(7)), out
        CA[6:0] values over 2 subsequent DRAM DDR clock edges.
    wck_sync : Signal(2), in
        One of `WCKSyncType`, determines WCK sync bit in CAS.
    """

    TRUTH_TABLE = {
        "DES":   "X X X X X X X             | X X X X X X X",       # DESELECT
        "NOP":   "L L L L L L L             | X X X X X X X",       # NO OPERATION
        "PDE":   "L L L L L L H             | X X X X X X X",       # POWER DOWN
        "ACT-1": "H H H R14-17              | BA0-3 R11-13",        # ACTIVATE-1
        "ACT-2": "H H L R7-10               | R0-6",                # ACTIVATE-2
        "PRE":   "L L L H H H H             | BA0-3 V V AB",        # PRECHARGE
        "REF":   "L L L H H H L             | BA0-2 RFM SB0 V AB",  # REFRESH
        "MWR":   "L H L C0 C3-5             | BA0-3 C1-2 AP",       # MASK WRITE
        "WR16":  "L H H C0 C3-5             | BA0-3 C1-2 AP",       # WRITE
        "WR32":  "L L H L C3-5              | BA0-3 C1-2 AP",       # WRITE32
        "RD16":  "H L L C0 C3-5             | BA0-3 C1-2 AP",       # READ
        "RD32":  "H L H C0 C3-5             | BA0-3 C1-2 AP",       # READ32
        "CAS":   "L L H H WS_WR WS_RD WS_FS | DC0-3 WRX WXSA WXSB", # CAS
        "MPC":   "L L L L H H OP7           | OP0-6",               # MULTI PURPOSE COMMAND
        "SRE":   "L L L H L H H             | V V V V V DSM PD",    # SELF REFRESH ENTRY
        "SRX":   "L L L H L H L             | V V V V V V V",       # SELF REFRESH EXIT
        "MRW-1": "L L L H H L H             | MA0-6",               # MODE REGISTER WRITE-1
        "MRW-2": "L L L H L L OP7           | OP0-6",               # MODE REGISTER WRITE-2
        "MRR":   "L L L H H L L             | MA0-6",               # MODE REGISTER READ
        "WFF":   "L L L L L H H             | L L L L L L L",       # WRITE FIFO
        "RFF":   "L L L L L H L             | L L L L L L L",       # READ FIFO
        "RDC":   "L L L L H L H             | L L L L L L L",       # READ DQ CALIBRATION
    }

    def _parse_truth_table(self):
        # transform TRUTH_TABLE to a form: {name: (['H', 'R1', 'R2', ...], [...]), ...}
        tt = {}
        for cmd, desc in self.TRUTH_TABLE.items():
            edges = desc.strip().split("|")
            assert len(edges) == 2, (cmd, desc)
            edges = map(self._parse_ranges, edges)
            pos_edge, neg_edge = map(lambda e: e.strip().split(), edges)
            for e in (pos_edge, neg_edge):
                assert len(neg_edge) == 7, (cmd, desc)
            tt[cmd] = (pos_edge, neg_edge)
        return tt

    def _parse_ranges(self, string):
        def replace(match):
            name = match.group(1)
            start, end = map(int, (match.group(2), match.group(3)))
            return " ".join(f"{name}{num}" for num in range(start, end+1))

        pattern = re.compile(r"([A-Z]+)(\d+)-(\d+)")
        return pattern.sub(replace, string)

    def __init__(self, dfi_phase, wck_sync, bank_organization=BankOrganization.B16):
        if bank_organization != BankOrganization.B16:
            raise NotImplementedError(f"Unsupported: {bank_organization}")
        self.truth_table = self._parse_truth_table()
        self.cs = Signal()
        self.ca = Array([Signal(7), Signal(7)])
        self.wck_sync = wck_sync
        self.dfi = dfi_phase

    def set(self, cmd):
        ops = []
        for edge, bits in enumerate(self.truth_table[cmd]):
            for bit, bit_desc in enumerate(bits):
                ops.append(self.ca[edge][bit].eq(self.parse_bit(bit_desc, cmd)))
        if cmd != "DES":  # only DESELECT has CS low
            ops.append(self.cs.eq(1))
        return ops

    def parse_bit(self, bit, cmd_str):
        assert len(self.dfi.bank) >= 7, "At least 7 DFI addressbits needed for Mode Register address"
        assert len(self.dfi.address) >= 18, "At least 18 DFI addressbits needed for row address"

        cmd = dfi_cmd(self.dfi)

        is_mrw = cmd_str.startswith("MRW")
        is_mpc = cmd_str == "MPC"

        mr_address = self.dfi.bank if is_mrw else self.dfi.address

        mpc_op = Signal(8)
        self.comb += If(self.dfi.address == 0,
            mpc_op.eq(MPC.ZQC_LATCH)
        ).Else(
            mpc_op.eq(self.dfi.address)
        )
        op = mpc_op if is_mpc else self.dfi.address

        rules = {
            "H":       lambda: 1,  # high
            "L":       lambda: 0,  # low
            "V":       lambda: 0,  # defined logic
            "X":       lambda: 0,  # don't care
            "AB":      lambda: self.dfi.address[10],  # all banks
            "AP":      lambda: self.dfi.address[10],  # auto precharge
            "RFM":     lambda: 0,  # TODO: 1=RFM, 0=REF (Refresh Managemenent, only if r/o MR[27][0]=1, else always REF)
            "SB(\d+)": lambda i: 0,  # sub-bank selection related to RFM
            "WS_WR":   lambda: self.wck_sync == WCKSyncType.WR,  # Write WCK2CK SYNC
            "WS_RD":   lambda: self.wck_sync == WCKSyncType.RD,  # Read WCK2CK SYNC
            "WS_FS":   lambda: self.wck_sync == WCKSyncType.FS,  # FAST SYNC
            "DC(\d+)": lambda i: 0,  # Data Copy, unimplemented
            "WRX":     lambda: 0,  # Write X function, unimplemented
            "WXSA":    lambda: 0,  # Write X function, unimplemented
            "WXSB":    lambda: 0,  # Write X function, unimplemented
            "BA(\d+)": lambda i: self.dfi.bank[i],  # only BA0-2 is used, in BG/B16 modes we always refresh banks (x, x+8)
            "R(\d+)":  lambda i: self.dfi.address[i],  # row
            # LPDDR5 specs split the regular column address into C[5:0] "column address" and B[3:0] "burst address"
            "C(\d+)":  lambda i: self.dfi.address[i + 4],
            "MA(\d+)": lambda i: mr_address[i],  # mode register address
            "OP(\d+)": lambda i: op[i], # mode register value, or operand for MPC
        }

        for pattern, value in rules.items():
            m = re.match(pattern, bit)
            if m:
                args = [int(g) for g in m.groups()]
                return value(*args)
        raise ValueError(bit)



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2018-2021 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import math

from migen import *

from litex.soc.interconnect import stream

from litedram.common import LiteDRAMNativePort
from litedram.frontend import dma

# Helpers ------------------------------------------------------------------------------------------

def _inc(signal, modulo):
    if modulo == 2**len(signal):
        return signal.eq(signal + 1)
    else:
        return If(signal == (modulo - 1),
            signal.eq(0)
        ).Else(
            signal.eq(signal + 1)
        )

# LiteDRAMFIFOCtrl ---------------------------------------------------------------------------------

class _LiteDRAMFIFOCtrl(Module):
    def __init__(self, base, depth):
        self.base  = base
        self.depth = depth
        self.level = Signal(max=depth+1)

        # # #

        # To write buffer
        self.writable      = Signal()
        self.write_address = Signal(max=depth)

        # From write buffer
        self.write = Signal()

        # To read buffer
        self.readable     = Signal()
        self.read_address = Signal(max=depth)

        # From read buffer
        self.read = Signal()

        # # #

        produce = self.write_address
        consume = self.read_address

        self.sync += [
            If(self.write,
                _inc(produce, depth)
            ),
            If(self.read,
                _inc(consume, depth)
            ),
            self.level.eq(self.level + self.write - self.read),
        ]

        self.comb += [
            self.writable.eq(self.level < depth),
            self.readable.eq(self.level > 0)
        ]

# LiteDRAMFIFOWriter -------------------------------------------------------------------------------

class _LiteDRAMFIFOWriter(Module):
    def __init__(self, data_width, port, ctrl, fifo_depth=16):
        self.sink = sink = stream.Endpoint([("data", data_width)])

        # # #

        self.submodules.writer = writer = dma.LiteDRAMDMAWriter(port, fifo_depth=fifo_depth)
        self.comb += [
            writer.sink.valid.eq(sink.valid & ctrl.writable),
            writer.sink.address.eq(ctrl.base + ctrl.write_address),
            writer.sink.data.eq(sink.data),
            If(writer.sink.valid & writer.sink.ready,
                sink.ready.eq(1),
                ctrl.write.eq(1)
            ),
        ]

# LiteDRAMFIFOReader -------------------------------------------------------------------------------

class _LiteDRAMFIFOReader(Module):
    def __init__(self, data_width, port, ctrl, fifo_depth=16):
        self.source = source = stream.Endpoint([("data", data_width)])

        # # #

        self.submodules.reader = reader = dma.LiteDRAMDMAReader(port, fifo_depth=fifo_depth)
        self.comb += [
            reader.sink.valid.eq(ctrl.readable),
            reader.sink.address.eq(ctrl.base + ctrl.read_address),
            If(reader.sink.valid & reader.sink.ready,
                ctrl.read.eq(1)
            )
        ]
        self.comb += reader.source.connect(source)

# _LiteDRAMFIFO ------------------------------------------------------------------------------------

class _LiteDRAMFIFO(Module):
    """LiteDRAM frontend that allows to use DRAM as a FIFO"""
    def __init__(self, data_width, base, depth, write_port, read_port,
        writer_fifo_depth = 16,
        reader_fifo_depth = 16):
        assert isinstance(write_port, LiteDRAMNativePort)
        assert isinstance(read_port,  LiteDRAMNativePort)
        self.sink   = stream.Endpoint([("data", data_width)])
        self.source = stream.Endpoint([("data", data_width)])

        # # #

        self.submodules.ctrl   = _LiteDRAMFIFOCtrl(base, depth)
        self.submodules.writer = _LiteDRAMFIFOWriter(data_width, write_port, self.ctrl, writer_fifo_depth)
        self.submodules.reader = _LiteDRAMFIFOReader(data_width, read_port,  self.ctrl, reader_fifo_depth)
        self.comb += [
            self.sink.connect(self.writer.sink),
            self.reader.source.connect(self.source)
        ]

# LiteDRAMFIFO -------------------------------------------------------------------------------------

class LiteDRAMFIFO(Module):
    """LiteDRAM FIFO with optional/automatic Bypass.


       Description
       -----------

                             ┌──────────┐        ┌──────────┐
                        Sink │   Pre-   │ Bypass │   Post-  │ Source
                    ─────────►   FIFO   ├────────►   FIFO   ├───────►
                             └────┬─────┘        └─────▲────┘
                                  │                    │
                             ┌────▼─────┐        ┌─────┴────┐
                             │   Pre-   │        │   Post-  │
                             │Converter │        │Converter │
                             └────┬─────┘        └─────▲────┘
                                  │                    │
                                  │  ┌─────────────┐   │
                                  │  │    DRAM     │   │
                                  └──►    FIFO     ├───┘
                                     └──────┬──────┘
                                            │
                                            ▼
                                          DRAM

       The DRAM FIFO allows creation of very large FIFO with storage in DRAM. The data-width of the
       input/output streams is automatically adapted to the DRAM's data-width with the Pre/Post con-
       verters and the module switches seamlessly between 2 modes:
       - 1) Bypass mode.
       - 2) DRAM mode.

       1) The module is initialized in Bypass mode, connecting the its Sink to its Source.
       Backpressure from the Source is propagated from the Source to the Post-FIFO, Pre-FIFO
       and the Sink.

                              ┌──────────┐        ┌──────────┐
                         Sink │   Pre-   │ Bypass │   Post-  │ Source
                     ─────────►   FIFO   ├────────►   FIFO   ├───────►
                              └──────────┘        └──────────┘
                                       Backpressure
                                  ◄─────────────────────

        Once the Post-FIFO is full and the Pre-FIFO has enough data to form a DRAM Word, the module
        switches to DRAM mode.

        2) In DRAM mode, the Bypass connection is disabled and Pre-FIFO's Source is redirected to
        Pre-Converter's Sink. Once Pre-Converter has a full DRAM word, the word can be written to the
        DRAM FIFO's Sink


                             ┌──────────┐        ┌──────────┐
                        Sink │   Pre-   │        │   Post-  │ Source
                    ─────────►   FIFO   │        │   FIFO   ├───────►
                             └────┬─────┘        └─────▲────┘
                                  │                    │
                             ┌────▼─────┐        ┌─────┴────┐
                             │   Pre-   │        │   Post-  │
                             │Converter │        │Converter │
                             └────┬─────┘        └─────▲────┘
                                  │                    │
                                  │  ┌─────────────┐   │
                                  │  │    DRAM     │   │
                                  └──►    FIFO     ├───┘
                                     └──────┬──────┘
                                            │
                                            ▼
                                          DRAM

        This data from DRAM FIFO will be generated back on the DRAM FIFO's Source and connected to
        the Post-Converter to re-generate the data with the correct data-width. Data will then be
        generated on the Source.

        Once we no longer have data in the Pre-Converter/DRAM FIFO/Post-Converter path and Pre-FIFO's
        level is below threshold, the modules switches back to Bypass mode.

    Parameters
    ----------
    data_width : int, in
        FIFO data-width.
    base : int, in
        FIFO base address in DRAM (bytes).
    depth: in, in
        FIFO depth (bytes).
    write_port: LiteDRAMNativePort
        DRAM Write port.
    read_port: LiteDRAMNativePort
        DRAM Read port.
    with_bypass: bool, in
        Automatic Bypass Mode Enable.
    """
    def __init__(self, data_width, base, depth, write_port, read_port, with_bypass=False,
        pre_fifo_depth  = 16,
        post_fifo_depth = 16):
        assert isinstance(write_port, LiteDRAMNativePort)
        assert isinstance(read_port,  LiteDRAMNativePort)
        self.sink   = stream.Endpoint([("data", data_width)])
        self.source = stream.Endpoint([("data", data_width)])

        # # #

        # Parameters.
        # -----------
        assert write_port.data_width == read_port.data_width
        port_data_width    = write_port.data_width
        port_address_width = write_port.address_width
        assert data_width <= port_data_width
        data_width_ratio = port_data_width//data_width
        if not with_bypass:
            assert data_width_ratio == 1
        fifo_base       = int(base/(port_data_width/8))
        fifo_depth      = int(depth/(port_data_width/8))
        pre_fifo_depth  = max( pre_fifo_depth, 2*data_width_ratio)
        post_fifo_depth = max(post_fifo_depth, 2*data_width_ratio)

        # Submodules.
        # -----------
        # Pre-FIFO.
        self.submodules.pre_fifo = pre_fifo = stream.SyncFIFO([("data", data_width)], pre_fifo_depth)

        # Pre-Converter.
        self.submodules.pre_converter = pre_converter = stream.Converter(data_width, port_data_width)

        # DRAM-FIFO.
        self.submodules.dram_fifo = dram_fifo = _LiteDRAMFIFO(
            data_width = port_data_width,
            base       = fifo_base,
            depth      = fifo_depth,
            write_port = write_port,
            read_port  = read_port,
        )

        # Post-Converter.
        self.submodules.post_converter = post_converter = stream.Converter(port_data_width, data_width)

        # Post-FIFO.
        self.submodules.post_fifo = post_fifo = stream.SyncFIFO([("data", data_width)], post_fifo_depth)

        # Data-Flow.
        # ----------
        dram_bypass          = Signal()
        dram_store           = Signal()
        dram_store_threshold = Signal()
        self.comb += [
            # Sink --> Pre-FIFO.
            self.sink.connect(pre_fifo.sink),

            # DRAM Threshold. We can only enable path to DRAM when we have enough data for a full
            # DRAM word.
            dram_store_threshold.eq(pre_fifo.level >= data_width_ratio),

            # Bypass / DRAM.
            If(with_bypass & dram_bypass,
                # Pre-FIFO --> Post-FIFO.
                pre_fifo.source.connect(post_fifo.sink),
            ).Else(
                # Pre-FIFO --> Pre-Converter.
                If(dram_store | (not with_bypass),
                    pre_fifo.source.connect(pre_converter.sink),
                ),
                # Post-Converter --> Post-FIFO.
                post_converter.source.connect(post_fifo.sink)
            ),

            # Pre-Converter --> DRAM-FIFO.
            pre_converter.source.connect(dram_fifo.sink),

            # DRAM-FIFO --> Post-Converter.
            dram_fifo.source.connect(post_converter.sink),

            # Post-FIFO --> Source.
            post_fifo.source.connect(self.source)
        ]

        # FSM.
        # ----
        if with_bypass:
            dram_first   = Signal()
            dram_inc     = Signal()
            dram_dec     = Signal()
            dram_cnt     = Signal(port_address_width)
            dram_inc_mod = Signal(max(int(math.log2(data_width_ratio)), 1))
            dram_dec_mod = Signal(max(int(math.log2(data_width_ratio)), 1))

            self.submodules.fsm = fsm = FSM(reset_state="BYPASS")
            fsm.act("BYPASS",
                dram_bypass.eq(1),
                # Switch to DRAM mode when enough data to store a DRAM word.
                If(dram_store_threshold,
                    NextValue(dram_first, 1),
                    NextValue(dram_cnt,   0),
                    NextState("DRAM")
                )
            )
            fsm.act("DRAM",
                # Store in DRAM.
                dram_store.eq(1),

                # Increment DRAM Data Count on Pre-Converter's Sink cycle.
                If(pre_converter.sink.valid & pre_converter.sink.ready,
                    dram_inc.eq(1),
                    NextValue(dram_first, 0),
                    If(data_width_ratio > 1,
                        NextValue(dram_inc_mod, dram_inc_mod + 1),
                    )
                ),

                # Decrement DRAM Data Count on Post-Converter's Source cycle.
                If(post_converter.source.valid & post_converter.source.ready,
                    dram_dec.eq(1),
                    If(data_width_ratio > 1,
                        NextValue(dram_dec_mod, dram_dec_mod + 1),
                    )
                ),

                # Maintain DRAM Data Count.
                NextValue(dram_cnt, dram_cnt + dram_inc - dram_dec),

                # Switch back to Bypass mode when no remaining DRAM word.
                If((dram_first == 0) & (dram_cnt == 0) & (dram_inc_mod == 0) & (dram_dec_mod == 0),
                    dram_store.eq(0),
                    NextState("BYPASS")
                )
            )



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2019 David Shah <dave@ds0.me>
# Copyright (c) 2019-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

# 1:2 frequency-ratio DDR3 PHY for Lattice's ECP5
# DDR3: 800 MT/s

from functools import reduce
from operator import or_

import math

from migen import *
from migen.genlib.misc import timeline
from migen.fhdl.specials import Tristate
from migen.genlib.cdc import MultiReg
from migen.genlib.misc import WaitTimer

from litex.soc.interconnect.csr import *

from litedram.common import *
from litedram.phy.dfi import *

# BitSlip ------------------------------------------------------------------------------------------

# FIXME: Use BitSlip from litedram.common.

class BitSlip(Module):
    def __init__(self, dw, rst=None, slp=None, cycles=1):
        self.i = Signal(dw)
        self.o = Signal(dw)
        self.rst = Signal() if rst is None else rst
        self.slp = Signal() if slp is None else slp

        # # #

        value = Signal(max=cycles*dw)
        self.sync += If(self.slp, value.eq(value + 1))
        self.sync += If(self.rst, value.eq(0))

        r = Signal((cycles+1)*dw, reset_less=True)
        self.sync += r.eq(Cat(r[dw:], self.i))
        cases = {}
        for i in range(cycles*dw):
            cases[i] = self.o.eq(r[i:dw+i])
        self.comb += Case(value, cases)

# Lattice ECP5 DDR PHY Initialization --------------------------------------------------------------

class ECP5DDRPHYInit(Module):
    def __init__(self):
        self.pause = Signal()
        self.stop  = Signal()
        self.delay = Signal()
        self.reset = Signal()

        # # #

        new_lock = Signal()
        update   = Signal()
        stop     = Signal()
        freeze   = Signal()
        pause    = Signal()
        reset    = Signal()

        # DDRDLLA instance -------------------------------------------------------------------------
        _lock = Signal()
        delay = Signal()
        self.specials += Instance("DDRDLLA",
            i_RST      = ResetSignal("init"),
            i_CLK      = ClockSignal("sys2x"),
            i_UDDCNTLN = ~update,
            i_FREEZE   = freeze,
            o_DDRDEL   = delay,
            o_LOCK     = _lock
        )
        lock   = Signal()
        lock_d = Signal()
        self.specials += MultiReg(_lock, lock, "init")
        self.sync.init += lock_d.eq(lock)
        self.comb += new_lock.eq(lock & ~lock_d)

        # DDRDLLA/DDQBUFM/ECLK initialization sequence ---------------------------------------------
        t = 8 # in cycles
        self.sync.init += [
            # Wait DDRDLLA Lock
            timeline(new_lock, [
                ( 1*t, [freeze.eq(1)]), # Freeze DDRDLLA
                ( 2*t, [  stop.eq(1)]), # Stop ECLK domain
                ( 3*t, [ reset.eq(1)]), # Reset ECLK domain
                ( 4*t, [ reset.eq(0)]), # Release ECLK domain reset
                ( 5*t, [  stop.eq(0)]), # Release ECLK domain stop
                ( 6*t, [freeze.eq(0)]), # Release DDRDLLA freeze
                ( 7*t, [ pause.eq(1)]), # Pause DQSBUFM
                ( 8*t, [update.eq(1)]), # Update DDRDLLA
                ( 9*t, [update.eq(0)]), # Release DDRDMMA update
                (10*t, [ pause.eq(0)]), # Release DQSBUFM pause
            ])
        ]

        # ------------------------------------------------------------------------------------------
        self.comb += [
            self.pause.eq(pause),
            self.stop.eq(stop),
            self.delay.eq(delay),
            self.reset.eq(reset),
        ]

# Lattice ECP5 DDR PHY -----------------------------------------------------------------------------

class ECP5DDRPHY(Module, AutoCSR):
    def __init__(self, pads,
        sys_clk_freq = 100e6,
        cl           = None,
        cwl          = None,
        cmd_delay    = 0,
        clk_polarity = 0,
        dm_remapping = None):
        assert isinstance(cmd_delay, int) and cmd_delay < 128
        pads        = PHYPadsCombiner(pads)
        memtype     = "DDR3"
        tck         = 2/(2*2*sys_clk_freq)
        addressbits = len(pads.a)
        bankbits    = len(pads.ba)
        nranks      = 1 if not hasattr(pads, "cs_n") else len(pads.cs_n)
        databits    = len(pads.dq)
        nphases     = 2
        if not dm_remapping:
            dm_remapping = {}
        assert databits%8 == 0

        # Init -------------------------------------------------------------------------------------
        self.submodules.init = ECP5DDRPHYInit()

        # Parameters -------------------------------------------------------------------------------
        cl  = get_default_cl( memtype, tck) if cl  is None else cl
        cwl = get_default_cwl(memtype, tck) if cwl is None else cwl
        cl_sys_latency  = get_sys_latency(nphases, cl)
        cwl_sys_latency = get_sys_latency(nphases, cwl)

        # Registers --------------------------------------------------------------------------------
        self._dly_sel = CSRStorage(databits//8)

        self._rdly_dq_rst         = CSR()
        self._rdly_dq_inc         = CSR()
        self._rdly_dq_bitslip_rst = CSR()
        self._rdly_dq_bitslip     = CSR()

        self._burstdet_clr  = CSR()
        self._burstdet_seen = CSRStatus(databits//8)

        # Observation
        self.datavalid = Signal(databits//8)

        # PHY settings -----------------------------------------------------------------------------
        rdphase = get_sys_phase(nphases, cl_sys_latency, cl)
        wrphase = get_sys_phase(nphases, cwl_sys_latency, cwl)
        self.settings = PhySettings(
            phytype       = "ECP5DDRPHY",
            memtype       = memtype,
            databits      = databits,
            dfi_databits  = 4*databits,
            nranks        = nranks,
            nphases       = nphases,
            rdphase       = rdphase,
            wrphase       = wrphase,
            cl            = cl,
            cwl           = cwl,
            read_latency  = cl_sys_latency + 10,
            write_latency = cwl_sys_latency,
            read_leveling = True,
            bitslips      = 4,
            delays        = 8,
        )

        # DFI Interface ----------------------------------------------------------------------------
        self.dfi = dfi = Interface(addressbits, bankbits, nranks, 4*databits, nphases)

        # # #

        bl8_chunk   = Signal()

        # Iterate on pads groups -------------------------------------------------------------------
        for pads_group in range(len(pads.groups)):
            pads.sel_group(pads_group)

            # Clock --------------------------------------------------------------------------------
            clk_pattern = {0: 0b1010, 1: 0b0101}[clk_polarity]
            for i in range(len(pads.clk_p)):
                pad_oddrx2f = Signal()
                self.specials += Instance("ODDRX2F",
                    i_RST  = ResetSignal("sys"),
                    i_SCLK = ClockSignal("sys"),
                    i_ECLK = ClockSignal("sys2x"),
                    **{f"i_D{n}": (clk_pattern >> n) & 0b1 for n in range(4)},
                    o_Q    = pad_oddrx2f
                )
                self.specials += Instance("DELAYG",
                    p_DEL_VALUE = cmd_delay,
                    i_A         = pad_oddrx2f,
                    o_Z         = pads.clk_p[i]
                )

            # Commands -----------------------------------------------------------------------------
            commands = {
                # Pad name: (DFI name,   Pad type (required or optional))
                "reset_n" : ("reset_n", "optional"),
                "cs_n"    : ("cs_n",    "optional"),
                "a"       : ("address", "required"),
                "ba"      : ("bank"   , "required"),
                "ras_n"   : ("ras_n"  , "required"),
                "cas_n"   : ("cas_n"  , "required"),
                "we_n"    : ("we_n"   , "required"),
                "cke"     : ("cke"    , "optional"),
                "odt"     : ("odt"    , "optional"),
            }
            for pad_name, (dfi_name, pad_type) in commands.items():
                pad = getattr(pads, pad_name, None)
                if (pad is None):
                    if (pad_type == "required"):
                        raise ValueError(f"DRAM pad {pad_name} required but not found in pads.")
                    continue
                for i in range(len(pad)):
                    pad_oddrx2f = Signal()
                    self.specials += Instance("ODDRX2F",
                        i_RST  = ResetSignal("sys"),
                        i_SCLK = ClockSignal("sys"),
                        i_ECLK = ClockSignal("sys2x"),
                        **{f"i_D{n}": getattr(dfi.phases[n//2], dfi_name)[i] for n in range(4)},
                        o_Q    = pad_oddrx2f
                    )
                    self.specials += Instance("DELAYG",
                        p_DEL_VALUE = cmd_delay,
                        i_A         = pad_oddrx2f,
                        o_Z         = pad[i]
                    )

        # DQS/DM/DQ --------------------------------------------------------------------------------
        dq_oe         = Signal()
        dqs_re        = Signal()
        dqs_oe        = Signal()
        dqs_postamble = Signal()
        dqs_preamble  = Signal()
        for i in range(databits//8):
            # DQSBUFM
            dqs_i    = Signal()
            dqsr90   = Signal()
            dqsw270  = Signal()
            dqsw     = Signal()
            rdpntr   = Signal(3)
            wrpntr   = Signal(3)
            rdly     = Signal(3)
            burstdet = Signal()
            self.sync += [
                If(self._dly_sel.storage[i] & self._rdly_dq_rst.re, rdly.eq(0)),
                If(self._dly_sel.storage[i] & self._rdly_dq_inc.re, rdly.eq(rdly + 1))
            ]
            self.specials += Instance("DQSBUFM",
                p_DQS_LI_DEL_ADJ = "MINUS",
                p_DQS_LI_DEL_VAL = 1,
                p_DQS_LO_DEL_ADJ = "MINUS",
                p_DQS_LO_DEL_VAL = 4,
                # Clocks / Reset
                i_RST            = ResetSignal("sys"),
                i_SCLK           = ClockSignal("sys"),
                i_ECLK           = ClockSignal("sys2x"),
                i_DDRDEL         = self.init.delay,
                i_PAUSE          = self.init.pause | self._dly_sel.storage[i],

                # Control
                # Assert LOADNs to use DDRDEL control
                i_RDLOADN        = 0,
                i_RDMOVE         = 0,
                i_RDDIRECTION    = 1,
                i_WRLOADN        = 0,
                i_WRMOVE         = 0,
                i_WRDIRECTION    = 1,

                # Reads (generate shifted DQS clock for reads)
                i_READ0          = dqs_re,
                i_READ1          = dqs_re,
                **{f"i_READCLKSEL{n}": rdly[n] for n in range(3)},
                i_DQSI           = dqs_i,
                o_DQSR90         = dqsr90,
                **{f"o_RDPNTR{n}": rdpntr[n] for n in range(3)},
                **{f"o_WRPNTR{n}": wrpntr[n] for n in range(3)},
                o_DATAVALID      = self.datavalid[i],
                o_BURSTDET       = burstdet,

                # Writes (generate shifted ECLK clock for writes)
                o_DQSW270        = dqsw270,
                o_DQSW           = dqsw
            )
            burstdet_d = Signal()
            self.sync += [
                burstdet_d.eq(burstdet),
                If(self._burstdet_clr.re,  self._burstdet_seen.status[i].eq(0)),
                If(burstdet & ~burstdet_d, self._burstdet_seen.status[i].eq(1)),
            ]

            # DQS ----------------------------------------------------------------------------------
            dqs      = Signal()
            dqs_oe_n = Signal()
            self.specials += [
                Instance("ODDRX2DQSB",
                    i_RST  = ResetSignal("sys"),
                    i_SCLK = ClockSignal("sys"),
                    i_ECLK = ClockSignal("sys2x"),
                    i_DQSW = dqsw,
                    **{f"i_D{n}": (0b1010 >> n) & 0b1 for n in range(4)},
                    o_Q    = dqs
                ),
                Instance("TSHX2DQSA",
                    i_RST  = ResetSignal("sys"),
                    i_SCLK = ClockSignal("sys"),
                    i_ECLK = ClockSignal("sys2x"),
                    i_DQSW = dqsw,
                    i_T0   = ~(dqs_oe | dqs_postamble),
                    i_T1   = ~(dqs_oe | dqs_preamble),
                    o_Q    = dqs_oe_n
                ),
                Tristate(pads.dqs_p[i], dqs, ~dqs_oe_n, dqs_i)
            ]

            # DM -----------------------------------------------------------------------------------
            dm_o_data       = Signal(8)
            dm_o_data_d     = Signal(8)
            dm_o_data_muxed = Signal(4)
            for n in range(8):
                self.comb += dm_o_data[n].eq(dfi.phases[n//4].wrdata_mask[n%4*databits//8+dm_remapping.get(i, i)])
            self.sync += dm_o_data_d.eq(dm_o_data)
            dm_bl8_cases = {}
            dm_bl8_cases[0] = dm_o_data_muxed.eq(dm_o_data[:4])
            dm_bl8_cases[1] = dm_o_data_muxed.eq(dm_o_data_d[4:])
            self.sync += Case(bl8_chunk, dm_bl8_cases)
            self.specials += Instance("ODDRX2DQA",
                i_RST     = ResetSignal("sys"),
                i_SCLK    = ClockSignal("sys"),
                i_ECLK    = ClockSignal("sys2x"),
                i_DQSW270 = dqsw270,
                **{f"i_D{n}": dm_o_data_muxed[n] for n in range(4)},
                o_Q       = pads.dm[i]
            )

            # DQ -----------------------------------------------------------------------------------
            for j in range(8*i, 8*(i+1)):
                dq_o            = Signal()
                dq_i            = Signal()
                dq_oe_n         = Signal()
                dq_i_delayed    = Signal()
                dq_i_data       = Signal(8)
                dq_o_data       = Signal(8)
                dq_o_data_d     = Signal(8)
                dq_o_data_muxed = Signal(4)
                for n in range(8):
                    self.comb += dq_o_data[n].eq(dfi.phases[n//4].wrdata[n%4*databits+j])
                self.sync += dq_o_data_d.eq(dq_o_data)
                dq_bl8_cases = {}
                dq_bl8_cases[0] = dq_o_data_muxed.eq(dq_o_data[:4])
                dq_bl8_cases[1] = dq_o_data_muxed.eq(dq_o_data_d[4:])
                self.sync += Case(bl8_chunk, dq_bl8_cases)
                self.specials += [
                    Instance("ODDRX2DQA",
                        i_RST     = ResetSignal("sys"),
                        i_SCLK    = ClockSignal("sys"),
                        i_ECLK    = ClockSignal("sys2x"),
                        i_DQSW270 = dqsw270,
                        **{f"i_D{n}": dq_o_data_muxed[n] for n in range(4)},
                        o_Q       = dq_o
                    )
                ]
                dq_i_bitslip = BitSlip(4,
                    rst    = self._dly_sel.storage[i] & self._rdly_dq_bitslip_rst.re,
                    slp    = self._dly_sel.storage[i] & self._rdly_dq_bitslip.re,
                    cycles = 1)
                self.submodules += dq_i_bitslip
                self.specials += [
                    Instance("DELAYG",
                        p_DEL_MODE = "DQS_ALIGNED_X2",
                        i_A        = dq_i,
                        o_Z        = dq_i_delayed
                    ),
                    Instance("IDDRX2DQA",
                        i_RST     = ResetSignal("sys"),
                        i_SCLK    = ClockSignal("sys"),
                        i_ECLK    = ClockSignal("sys2x"),
                        i_DQSR90  = dqsr90,
                        **{f"i_RDPNTR{n}": rdpntr[n] for n in range(3)},
                        **{f"i_WRPNTR{n}": wrpntr[n] for n in range(3)},
                        i_D       = dq_i_delayed,
                        **{f"o_Q{n}": dq_i_bitslip.i[n] for n in range(4)},
                    )
                ]
                dq_i_bitslip_o_d = Signal(4)
                self.sync += dq_i_bitslip_o_d.eq(dq_i_bitslip.o)
                self.comb += dq_i_data.eq(Cat(dq_i_bitslip_o_d, dq_i_bitslip.o))
                for n in range(8):
                    self.comb += dfi.phases[n//4].rddata[n%4*databits+j].eq(dq_i_data[n])
                self.specials += [
                    Instance("TSHX2DQA",
                        i_RST     = ResetSignal("sys"),
                        i_SCLK    = ClockSignal("sys"),
                        i_ECLK    = ClockSignal("sys2x"),
                        i_DQSW270 = dqsw270,
                        i_T0      = ~dq_oe,
                        i_T1      = ~dq_oe,
                        o_Q       = dq_oe_n,
                    ),
                    Tristate(pads.dq[j], dq_o, ~dq_oe_n, dq_i)
                ]

        # Read Control Path ------------------------------------------------------------------------
        rdtap = cl_sys_latency

        # Creates a delay line of read commands coming from the DFI interface. The taps are used to
        # control DQS read (internal read pulse of the DQSBUF) and the output of the delay is used
        # signal a valid read data to the DFI interface.
        #
        # The DQS read must be asserted for 2 sys_clk cycles before the read data is coming back from
        # the DRAM (see 6.2.4 READ Pulse Positioning Optimization of FPGA-TN-02035-1.2)
        #
        # The read data valid is asserted for 1 sys_clk cycle when the data is available on the DFI
        # interface, the latency is the sum of the ODDRX2DQA, CAS, IDDRX2DQA latencies.
        rddata_en = TappedDelayLine(
            signal = reduce(or_, [dfi.phases[i].rddata_en for i in range(nphases)]),
            ntaps  = self.settings.read_latency
        )
        self.submodules += rddata_en

        self.comb += [phase.rddata_valid.eq(rddata_en.output) for phase in dfi.phases]
        self.comb += dqs_re.eq(rddata_en.taps[rdtap] | rddata_en.taps[rdtap + 1])

        # Write Control Path -----------------------------------------------------------------------
        wrtap = cwl_sys_latency

        # Create a delay line of write commands coming from the DFI interface. This taps are used to
        # control DQ/DQS tristates and to select write data of the DRAM burst from the DFI interface.
        # The PHY is operating in halfrate mode (so provide 4 datas every sys_clk cycles: 2x for DDR,
        # 2x for halfrate) but DDR3 requires a burst of 8 datas (BL8) for best efficiency. Writes are
        # then performed in 2 sys_clk cycles and data needs to be selected for each cycle.
        wrdata_en = TappedDelayLine(
            signal = reduce(or_, [dfi.phases[i].wrdata_en for i in range(nphases)]),
            ntaps  = wrtap + 4
        )
        self.submodules += wrdata_en

        self.comb += dq_oe.eq(wrdata_en.taps[wrtap] | wrdata_en.taps[wrtap + 1])
        self.comb += bl8_chunk.eq(wrdata_en.taps[wrtap])
        self.comb += dqs_oe.eq(dq_oe)

        # Write DQS Postamble/Preamble Control Path ------------------------------------------------
        # Generates DQS Preamble 1 cycle before the first write and Postamble 1 cycle after the last
        # write. During writes, DQS tristate is configured as output for at least 4 sys_clk cycles:
        # 1 for Preamble, 2 for the Write and 1 for the Postamble.
        self.comb += dqs_preamble.eq( wrdata_en.taps[wrtap - 1]  & ~wrdata_en.taps[wrtap + 0])
        self.comb += dqs_postamble.eq(wrdata_en.taps[wrtap + 2]  & ~wrdata_en.taps[wrtap + 1])



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2016-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.interconnect import stream

from litedram.common import *

# LiteDRAMNativePortCDC ----------------------------------------------------------------------------

class LiteDRAMNativePortCDC(Module):
    def __init__(self, port_from, port_to,
                 cmd_depth   = 4,
                 wdata_depth = 16,
                 rdata_depth = 16):
        assert port_from.address_width == port_to.address_width
        assert port_from.data_width    == port_to.data_width
        assert port_from.mode          == port_to.mode

        address_width = port_from.address_width
        data_width    = port_from.data_width
        mode          = port_from.mode

        # # #

        cmd_cdc = stream.ClockDomainCrossing(
            layout  = [("we", 1), ("addr", address_width)],
            cd_from = port_from.clock_domain,
            cd_to   = port_to.clock_domain,
            depth   = cmd_depth,
            with_common_rst = False)
        self.submodules += cmd_cdc
        self.submodules += stream.Pipeline(port_from.cmd, cmd_cdc, port_to.cmd)

        if mode in ["write", "both"]:
            wdata_cdc = stream.ClockDomainCrossing(
                layout  = [("data", data_width), ("we", data_width//8)],
                cd_from = port_from.clock_domain,
                cd_to   = port_to.clock_domain,
                depth   = wdata_depth,
                with_common_rst = False)
            self.submodules += wdata_cdc
            self.submodules += stream.Pipeline(port_from.wdata, wdata_cdc, port_to.wdata)

        if mode in ["read", "both"]:
            rdata_cdc = stream.ClockDomainCrossing(
                layout  = [("data", data_width)],
                cd_from = port_to.clock_domain,
                cd_to   = port_from.clock_domain,
                depth   = rdata_depth,
                with_common_rst = False)
            self.submodules += rdata_cdc
            self.submodules += stream.Pipeline(port_to.rdata, rdata_cdc, port_from.rdata)

# LiteDRAMNativePortDownConverter ------------------------------------------------------------------

class LiteDRAMNativePortDownConverter(Module):
    """LiteDRAM port DownConverter

    This module reduces user port data width to fit controller data width.
    With N = port_from.data_width/port_to.data_width:
    - Address is adapted (multiplied by N + internal increments)
    - A write from the user is splitted and generates N writes to the
    controller.
    - A read from the user generates N reads to the controller and returned
      datas are regrouped in a single data presented to the user.
    """
    def __init__(self, port_from, port_to, reverse=False):
        assert port_from.clock_domain == port_to.clock_domain
        assert port_from.data_width    > port_to.data_width
        assert port_from.mode         == port_to.mode
        if port_from.data_width % port_to.data_width:
            raise ValueError("Ratio must be an int")

        # # #

        ratio = port_from.data_width//port_to.data_width
        mode  = port_from.mode
        count = Signal(max=ratio)

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            NextValue(count, 0),
            If(port_from.cmd.valid,
                NextState("CONVERT")
            )
        )
        fsm.act("CONVERT",
            port_to.cmd.valid.eq(1),
            port_to.cmd.we.eq(port_from.cmd.we),
            port_to.cmd.addr.eq(port_from.cmd.addr*ratio + count),
            If(port_to.cmd.ready,
                NextValue(count, count + 1),
                If(count == (ratio - 1),
                    port_from.cmd.ready.eq(1),
                    NextState("IDLE")
                )
            )
        )

        if mode in ["write", "both"]:
            wdata_converter = stream.StrideConverter(
                description_from = port_from.wdata.description,
                description_to   = port_to.wdata.description,
                reverse          = reverse)
            self.submodules += wdata_converter
            self.submodules += stream.Pipeline(port_from.wdata, wdata_converter, port_to.wdata)

        if mode in ["read", "both"]:
            rdata_converter = stream.StrideConverter(
                description_from = port_to.rdata.description,
                description_to   = port_from.rdata.description,
                reverse          = reverse)
            self.submodules += rdata_converter
            self.submodules += stream.Pipeline(
                port_to.rdata, rdata_converter, port_from.rdata)

# LiteDRAMNativePortUpConverter --------------------------------------------------------------------

class LiteDRAMNativePortUpConverter(Module):
    """LiteDRAM port UpConverter

    This module increase user port data width to fit controller data width.
    With N = port_to.data_width/port_from.data_width:
    - Address is adapted (divided by N)
    - N read from user are regrouped in a single one to the controller
    (when possible, ie when consecutive and bursting)
    - N writes from user are regrouped in a single one to the controller
    (when possible, ie when consecutive and bursting)
    Incomplete writes/reads (i.e. with n < N) are handled automatically in the
    middle of a burst, but last command has to use cmd.last=1 if the last burst
    is not complete (not all N addresses have been used).
    """
    def __init__(self, port_from, port_to, reverse=False):
        assert port_from.clock_domain == port_to.clock_domain
        assert port_from.data_width    < port_to.data_width
        assert port_from.mode         == port_to.mode
        if port_to.data_width % port_from.data_width:
            raise ValueError("Ratio must be an int")

        # # #

        ratio = port_to.data_width//port_from.data_width
        mode  = port_from.mode

        # Command ----------------------------------------------------------------------------------

        # Defines cmd type and the chunks that have been requested for the current port_to command.
        sel              = Signal(ratio)
        cmd_buffer       = stream.SyncFIFO([("sel", ratio), ("we", 1)], 0)
        self.submodules += cmd_buffer
        # Store last received command.
        cmd_addr         = Signal.like(port_from.cmd.addr)
        cmd_we           = Signal()
        cmd_last         = Signal()
        # Indicates that we need to proceed to the next port_to command.
        next_cmd         = Signal()
        addr_changed     = Signal()
        # Signals that indicate that write/read convertion has finished.
        wdata_finished   = Signal()
        rdata_finished   = Signal()
        # Used to prevent reading old memory value if previous command has written the same address.
        read_lock        = Signal()
        read_unlocked    = Signal()
        rw_collision     = Signal()

        # Different order depending on read/write:
        # - read:  new -> cmd -> fill -> commit -> new
        # - write: new -> fill -> commit -> cmd -> new
        # For writes we have to send the command at the end to prevent situations when, during
        # a burst, LiteDRAM expects data (wdata_ready=1) but write converter is still converting.
        self.submodules.fsm = fsm = FSM()
        fsm.act("NEW",
            port_from.cmd.ready.eq(port_from.cmd.valid & ~read_lock),
            If(port_from.cmd.ready,
                NextValue(cmd_addr, port_from.cmd.addr),
                NextValue(cmd_we, port_from.cmd.we),
                NextValue(cmd_last, port_from.cmd.last),
                NextValue(sel, 1 << port_from.cmd.addr[:log2_int(ratio)]),
                If(port_from.cmd.we,
                    NextState("FILL"),
                ).Else(
                    NextState("CMD"),
                )
            )
        )
        fsm.act("CMD",
            port_to.cmd.valid.eq(1),
            port_to.cmd.we.eq(cmd_we),
            port_to.cmd.addr.eq(cmd_addr[log2_int(ratio):]),
            If(port_to.cmd.ready,
                If(cmd_we,
                    NextState("NEW")
                ).Else(
                    NextState("FILL")
                )
            )
        )
        fsm.act("FILL",
            If(next_cmd,
                NextState("COMMIT")
            ).Else(  # Acknowledge incomming commands, while filling `sel`.
                port_from.cmd.ready.eq(port_from.cmd.valid),
                NextValue(cmd_last, port_from.cmd.last),
                If(port_from.cmd.valid,
                    NextValue(sel, sel | 1 << port_from.cmd.addr[:log2_int(ratio)])
                )
            )
        )
        fsm.act("COMMIT",
            cmd_buffer.sink.valid.eq(1),
            cmd_buffer.sink.sel.eq(sel),
            cmd_buffer.sink.we.eq(cmd_we),
            If(cmd_buffer.sink.ready,
                If(cmd_we,
                    NextState("CMD")
                ).Else(
                    NextState("NEW")
                )
            )
        )

        self.comb += [
            cmd_buffer.source.ready.eq(wdata_finished | rdata_finished),
            addr_changed.eq(cmd_addr[log2_int(ratio):] != port_from.cmd.addr[log2_int(ratio):]),
            # Collision happens on write to read transition when address does not change.
            rw_collision.eq(cmd_we & (port_from.cmd.valid & ~port_from.cmd.we) & ~addr_changed),
            # Go to the next command if one of the following happens:
            #  - port_to address changes.
            #  - cmd type changes.
            #  - we received all the `ratio` commands.
            #  - this is the last command in a sequence.
            #  - master requests a flush (even after the command has been sent).
            next_cmd.eq(addr_changed | (cmd_we != port_from.cmd.we) | (sel == 2**ratio - 1)
                        | cmd_last | port_from.flush),
        ]

        self.sync += [
            # Block sending read command if we have just written to that address
            If(wdata_finished,
                read_lock.eq(0),
                read_unlocked.eq(1),
            ).Elif(rw_collision & ~port_to.cmd.valid & ~read_unlocked,
                read_lock.eq(1)
            ),
            If(port_from.cmd.valid & port_from.cmd.ready,
                read_unlocked.eq(0)
            )
        ]

        # Read Datapath ----------------------------------------------------------------------------

        if mode in ["read", "both"]:
            # Queue received data not to loose it when it comes too fast.
            rdata_fifo = stream.SyncFIFO(port_to.rdata.description, ratio - 1)
            rdata_converter = stream.StrideConverter(
                description_from = port_to.rdata.description,
                description_to   = port_from.rdata.description,
                reverse          = reverse)
            self.submodules +=  rdata_fifo, rdata_converter

            # Shift register with a bitmask of current chunk.
            rdata_chunk       = Signal(ratio, reset=1)
            rdata_chunk_valid = Signal()
            self.sync += \
                If(rdata_converter.source.valid &
                   rdata_converter.source.ready,
                    rdata_chunk.eq(Cat(rdata_chunk[ratio-1], rdata_chunk[:ratio-1]))
                )

            self.comb += [
                # port_to -> rdata_fifo -> rdata_converter -> port_from
                port_to.rdata.connect(rdata_fifo.sink),
                rdata_fifo.source.connect(rdata_converter.sink),
                rdata_chunk_valid.eq((cmd_buffer.source.sel & rdata_chunk) != 0),
                If(cmd_buffer.source.valid & ~cmd_buffer.source.we,
                   # If that chunk is valid we send it to the user port and wait for ready.
                    If(rdata_chunk_valid,
                        port_from.rdata.valid.eq(rdata_converter.source.valid),
                        port_from.rdata.data.eq(rdata_converter.source.data),
                        rdata_converter.source.ready.eq(port_from.rdata.ready)
                    ).Else(  # If this chunk was not requested in `sel`, ignore it.
                        rdata_converter.source.ready.eq(1)
                    ),
                    rdata_finished.eq(rdata_converter.source.valid & rdata_converter.source.ready
                                      & rdata_chunk[ratio - 1])
                ),
            ]

        # Write Datapath ---------------------------------------------------------------------------

        if mode in ["write", "both"]:
            # Queue write data not to miss it when the lower chunks haven't been reqested.
            wdata_fifo    = stream.SyncFIFO(port_from.wdata.description, ratio - 1)
            wdata_buffer  = stream.SyncFIFO(port_to.wdata.description, 1)
            wdata_converter = stream.StrideConverter(
                description_from = port_from.wdata.description,
                description_to   = port_to.wdata.description,
                reverse          = reverse)
            self.submodules += wdata_converter, wdata_fifo, wdata_buffer

            # Shift register with a bitmask of current chunk.
            wdata_chunk       = Signal(ratio, reset=1)
            wdata_chunk_valid = Signal()
            self.sync += \
                If(wdata_converter.sink.valid & wdata_converter.sink.ready,
                    wdata_chunk.eq(Cat(wdata_chunk[ratio-1], wdata_chunk[:ratio-1]))
                )

            # Replicate `sel` bits to match the width of port_to.wdata.we.
            wdata_sel = Signal.like(port_to.wdata.we)
            if reverse:
                wdata_sel_parts = [
                    Replicate(cmd_buffer.source.sel[i], port_to.wdata.we.nbits // sel.nbits)
                    for i in reversed(range(ratio))
                ]
            else:
                wdata_sel_parts = [
                    Replicate(cmd_buffer.source.sel[i], port_to.wdata.we.nbits // sel.nbits)
                    for i in range(ratio)
                ]

            self.sync += \
                If(cmd_buffer.source.valid & cmd_buffer.source.we & wdata_chunk[ratio - 1],
                    wdata_sel.eq(Cat(wdata_sel_parts))
                )

            self.comb += [
                # port_from -> wdata_fifo -> wdata_converter
                port_from.wdata.connect(wdata_fifo.sink),
                wdata_buffer.source.connect(port_to.wdata),
                wdata_chunk_valid.eq((cmd_buffer.source.sel & wdata_chunk) != 0),
                If(cmd_buffer.source.valid & cmd_buffer.source.we,
                    # When the current chunk is valid, read it from wdata_fifo.
                    If(wdata_chunk_valid,
                        wdata_converter.sink.valid.eq(wdata_fifo.source.valid),
                        wdata_converter.sink.data.eq(wdata_fifo.source.data),
                        wdata_converter.sink.we.eq(wdata_fifo.source.we),
                        wdata_fifo.source.ready.eq(wdata_converter.sink.ready),
                    ).Else(  # If chunk is not valid, send any data and do not advance fifo.
                        wdata_converter.sink.valid.eq(1),
                    ),
                ),
                wdata_buffer.sink.valid.eq(wdata_converter.source.valid),
                wdata_buffer.sink.data.eq(wdata_converter.source.data),
                wdata_buffer.sink.we.eq(wdata_converter.source.we & wdata_sel),
                wdata_converter.source.ready.eq(wdata_buffer.sink.ready),
                wdata_finished.eq(wdata_converter.sink.valid & wdata_converter.sink.ready
                                  & wdata_chunk[ratio-1]),
            ]

# LiteDRAMNativePortConverter ----------------------------------------------------------------------

class LiteDRAMNativePortConverter(Module):
    def __init__(self, port_from, port_to, reverse=False):
        assert port_from.clock_domain == port_to.clock_domain
        assert port_from.mode         == port_to.mode

        # # #

        ratio = port_from.data_width/port_to.data_width

        if ratio > 1:
            # DownConverter
            self.submodules.converter = LiteDRAMNativePortDownConverter(port_from, port_to, reverse)
        elif ratio < 1:
            # UpConverter
            self.submodules.converter = LiteDRAMNativePortUpConverter(port_from, port_to, reverse)
        else:
            # Identity
            self.comb += port_from.connect(port_to)



#!/usr/bin/env python3

from setuptools import setup
from setuptools import find_packages


with open("README.md", "r") as fp:
    long_description = fp.read()


setup(
    name="litedram",
    description="Small footprint and configurable DRAM core",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Florent Kermarrec",
    version="2022.12",
    author_email="florent@enjoy-digital.fr",
    url="http://enjoy-digital.fr",
    download_url="https://github.com/enjoy-digital/litedram",
    test_suite="test",
    license="BSD",
    python_requires="~=3.6",
    install_requires=["pyyaml", "litex"],
    extras_require={
        "develop": [
          "meson"
          "pexpect"
          "setuptools"
          "requests"
        ]
    },
    packages=find_packages(exclude=("test*", "sim*", "doc*", "examples*")),
    include_package_data=True,
    keywords="HDL ASIC FPGA hardware design",
    classifiers=[
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Environment :: Console",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    entry_points={
        "console_scripts": [
            "litedram_gen=litedram.gen:main",
        ],
    },
)



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2018-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

"""
AXI frontend for LiteDRAM

Converts AXI ports to Native ports.

Features:
- Write/Read arbitration.
- Write/Read data buffers (configurable depth).
- Burst support (FIXED/INCR/WRAP).
- ID support (configurable width).
- Optional Read-Modify-Write support (When only full words can be written on the DRAM, ex with ECC).

Limitations:
- Response always okay.
- No reordering.
"""

from migen import *
from migen.genlib.record import *
from migen.genlib.roundrobin import *

from litex.soc.interconnect import stream
from litex.soc.interconnect.axi import *

# LiteDRAMAXIPort ----------------------------------------------------------------------------------

class LiteDRAMAXIPort(AXIInterface): pass

# LiteDRAMAXI2NativeW ------------------------------------------------------------------------------

class LiteDRAMAXI2NativeW(Module):
    def __init__(self, axi, port, buffer_depth, base_address, with_read_modify_write=False):
        assert axi.address_width >= log2_int(base_address)
        assert axi.data_width    == port.data_width
        self.cmd_request = Signal()
        self.cmd_grant   = Signal()

        # # #

        can_write = Signal()

        ashift = log2_int(port.data_width//8)

        # Burst to Beat ----------------------------------------------------------------------------
        aw        =  AXIStreamInterface(layout=ax_description(axi.address_width), id_width=axi.id_width)
        aw_buffer = stream.Buffer(aw.description)
        self.submodules += aw_buffer
        self.comb += axi.aw.connect(aw_buffer.sink)
        aw_burst2beat = AXIBurst2Beat(aw_buffer.source, aw)
        self.submodules.aw_burst2beat = aw_burst2beat

        # Write Buffer -----------------------------------------------------------------------------
        w = AXIStreamInterface(layout=w_description(axi.data_width), id_width=axi.id_width)
        w_buffer = stream.SyncFIFO(w.description, depth=buffer_depth, buffered=True)
        self.submodules.w_buffer = w_buffer

        # Write ID Buffer & Response ---------------------------------------------------------------
        id_buffer   = stream.SyncFIFO([("id", axi.id_width)], buffer_depth)
        resp_buffer = stream.SyncFIFO([("id", axi.id_width), ("resp", 2)], buffer_depth)
        self.submodules += id_buffer, resp_buffer
        self.comb += [
            id_buffer.sink.valid.eq(aw.valid & aw.first & aw.ready),
            id_buffer.sink.id.eq(aw.id),
            If(w_buffer.source.valid &
               w_buffer.source.last &
               w_buffer.source.ready,
                resp_buffer.sink.valid.eq(1),
                resp_buffer.sink.resp.eq(RESP_OKAY),
                resp_buffer.sink.id.eq(id_buffer.source.id),
                id_buffer.source.ready.eq(1)
            ),
            resp_buffer.source.connect(axi.b, omit={"id"}), # FIXME: Avoid manual id connection.
            axi.b.id.eq(resp_buffer.source.id),             # FIXME: Avoid manual id connection.
        ]

        # Write Buffer reservation ------------------------------------------------------------------
        # - Incremented when data cmd is send
        # - Decremented when data is read
        w_buffer_queue   = Signal()
        w_buffer_dequeue = Signal()
        w_buffer_level   = Signal(max=buffer_depth + 1)
        self.comb += [
            w_buffer_queue.eq(port.cmd.valid & port.cmd.ready & port.cmd.we),
            w_buffer_dequeue.eq(w_buffer.source.valid & w_buffer.source.ready)
        ]
        self.sync += [
            If(w_buffer_queue,
                If(~w_buffer_dequeue, w_buffer_level.eq(w_buffer_level + 1))
            ).Elif(w_buffer_dequeue,
                w_buffer_level.eq(w_buffer_level - 1)
            )
        ]
        self.comb += can_write.eq(w_buffer.level > w_buffer_level)

        # Command ----------------------------------------------------------------------------------
        # Accept and send command to the controller only if:
        # - Address & Data request are *both* valid.
        # - Data buffer is not empty.
        self.comb += [
            self.cmd_request.eq(aw.valid & can_write),
            If(self.cmd_request & self.cmd_grant,
                port.cmd.valid.eq(1),
                port.cmd.last.eq(aw.last),
                port.cmd.we.eq(1),
                port.cmd.addr.eq((aw.addr - base_address) >> ashift),
                If(port.cmd.ready,
                    aw.ready.eq(1),
                )
            )
        ]

        # Write Data -------------------------------------------------------------------------------
        axi_w_connect = Signal(reset=1)
        self.comb += [
            If(axi_w_connect, axi.w.connect(w_buffer.sink)),
            w_buffer.source.connect(port.wdata, omit={"strb", "id", "dest", "user"}),
            port.wdata.we.eq(w_buffer.source.strb)
        ]

        # Read-Modify-Write ------------------------------------------------------------------------
        if with_read_modify_write:
            # RMW Request/Grant signals.
            self.rmw_request = Signal()
            self.rmw_rgrant  = Signal()
            self.rmw_wgrant  = Signal()

            # # #

            rmw_data      = Signal(port.data_width)
            rmw_mask      = Signal(port.data_width)
            rmw_cmd_done  = Signal()
            rmw_data_done = Signal()

            # Grant write when write buffer is empty.
            self.comb += self.rmw_wgrant.eq(~w_buffer_queue & (w_buffer_level == 0))

            # Prevent new write on Read-Modify-Write request.
            self.comb += If(self.rmw_request,
                can_write.eq(0)
            )

            # Disconnect regular Datapath on a Read-Modify-Write cycle.
            self.comb += If(self.rmw_request,
                axi_w_connect.eq(0),
            )

            # Read-Modify-Write FSM.
            self.submodules.rmw_fsm = rmw_fsm = FSM(reset_state="IDLE")
            rmw_fsm.act("IDLE",
                # Clear RMW Cmd/Data done signals.
                NextValue(rmw_cmd_done,  0),
                NextValue(rmw_data_done, 0),
                # Detect partial data and initiate a RMW access.
                If(axi.w.valid & (axi.w.strb != (2**len(axi.w.strb) - 1)),
                    # Before issuing the RMW sequence, we must ensure that all pending writes/reads
                    # access have been done, so issue a request and wait for grant.
                    self.rmw_request.eq(1),
                    If(self.rmw_rgrant & self.rmw_wgrant,
                        NextState("READ")
                    )
                )
            )
            rmw_fsm.act("READ",
                self.rmw_request.eq(1),
                # Issue Read Cmd.
                port.cmd.valid.eq(1),
                port.cmd.last.eq(aw.last),
                port.cmd.we.eq(0),
                port.cmd.addr.eq((aw.addr - base_address) >> ashift),
                If(port.cmd.ready,
                    NextState("MODIFY")
                )
            )
            rmw_fsm.act("MODIFY",
                self.rmw_request.eq(1),
                # Generate mask.
                *[rmw_mask[8*i:8*(i+1)].eq(Replicate(axi.w.strb[i], 8)) for i in range(port.data_width//8)],
                # Receive Read Data and modify it.
                port.rdata.ready.eq(1),
                If(port.rdata.valid,
                    # Keep previous unmasked data and replace masked data with new ones.
                    NextValue(rmw_data, (port.rdata.data & ~rmw_mask) | (axi.w.data & rmw_mask)),
                    NextState("WRITE")
                )
            )
            rmw_fsm.act("WRITE",
                self.rmw_request.eq(1),
                # Isssue Write Cmd.
                port.cmd.valid.eq(~rmw_cmd_done),
                port.cmd.last.eq(aw.last),
                port.cmd.we.eq(1),
                port.cmd.addr.eq((aw.addr - base_address) >> ashift),
                If(port.cmd.valid & port.cmd.ready,
                    aw.ready.eq(1),
                    NextValue(rmw_cmd_done, 1)
                ),
                # Issue Write Data.
                w_buffer.sink.valid.eq(~rmw_data_done),
                w_buffer.sink.last.eq(axi.w.last),
                w_buffer.sink.data.eq(rmw_data),
                w_buffer.sink.strb.eq(2**len(w_buffer.sink.strb) - 1),
                If(w_buffer.sink.valid & w_buffer.sink.ready,
                    axi.w.ready.eq(1),
                    NextValue(rmw_data_done, 1)
                ),
                # Return to Idle when both Cmd/Data are done.
                If((port.cmd.ready | rmw_cmd_done) & (w_buffer.sink.ready | rmw_data_done),
                    NextState("IDLE")
                )
            )

# LiteDRAMAXI2NativeR ------------------------------------------------------------------------------

class LiteDRAMAXI2NativeR(Module):
    def __init__(self, axi, port, buffer_depth, base_address, with_read_modify_write=False):
        assert axi.address_width >= log2_int(base_address)
        assert axi.data_width    == port.data_width
        self.cmd_request = Signal()
        self.cmd_grant   = Signal()

        # # #

        can_read = Signal()

        ashift = log2_int(port.data_width//8)

        # Burst to Beat ----------------------------------------------------------------------------
        ar        =  AXIStreamInterface(layout=ax_description(axi.address_width), id_width=axi.id_width)
        ar_buffer = stream.Buffer(ar.description)
        self.submodules += ar_buffer
        self.comb += axi.ar.connect(ar_buffer.sink)
        ar_burst2beat = AXIBurst2Beat(ar_buffer.source, ar)
        self.submodules.ar_burst2beat = ar_burst2beat

        # Read buffer ------------------------------------------------------------------------------
        r = AXIStreamInterface(layout=r_description(axi.data_width), id_width=axi.id_width)
        r_buffer = stream.SyncFIFO(r.description, depth=buffer_depth, buffered=True)
        self.submodules.r_buffer = r_buffer

        # Read Buffer reservation ------------------------------------------------------------------
        # - Incremented when data is planned to be queued
        # - Decremented when data is dequeued
        r_buffer_queue   = Signal()
        r_buffer_dequeue = Signal()
        r_buffer_level   = Signal(max=buffer_depth + 1)
        self.comb += [
            r_buffer_queue.eq(port.cmd.valid & port.cmd.ready & ~port.cmd.we),
            r_buffer_dequeue.eq(r_buffer.source.valid & r_buffer.source.ready)
        ]
        self.sync += [
            If(r_buffer_queue,
                If(~r_buffer_dequeue, r_buffer_level.eq(r_buffer_level + 1))
            ).Elif(r_buffer_dequeue,
                r_buffer_level.eq(r_buffer_level - 1)
            )
        ]
        self.comb += can_read.eq(r_buffer_level != buffer_depth)

        # Read ID Buffer ---------------------------------------------------------------------------
        id_buffer = stream.SyncFIFO([("id", axi.id_width)], buffer_depth)
        self.submodules += id_buffer
        self.comb += [
            id_buffer.sink.valid.eq(ar.valid & ar.ready),
            id_buffer.sink.last.eq(ar.last),
            id_buffer.sink.id.eq(ar.id),
            axi.r.last.eq(id_buffer.source.last),
            axi.r.id.eq(id_buffer.source.id),
            id_buffer.source.ready.eq(axi.r.valid & axi.r.ready)
        ]

        # Command ----------------------------------------------------------------------------------
        self.comb += [
            self.cmd_request.eq(ar.valid & can_read),
            If(self.cmd_request & self.cmd_grant,
                port.cmd.valid.eq(1),
                port.cmd.last.eq(ar.last),
                port.cmd.we.eq(0),
                port.cmd.addr.eq((ar.addr - base_address) >> ashift),
                If(port.cmd.ready,
                    ar.ready.eq(1),
                )
            )
        ]

        # Read data --------------------------------------------------------------------------------
        self.comb += [
            port.rdata.connect(r_buffer.sink, omit={"bank"}),
            r_buffer.source.connect(axi.r, omit={"id", "last"}),
            axi.r.resp.eq(RESP_OKAY)
        ]

        # Read-Modify-Write ------------------------------------------------------------------------
        if with_read_modify_write:
            # RMW Request/Grant signals.
            self.rmw_request = Signal()
            self.rmw_rgrant  = Signal()

            # # #

            # Grant read when read buffer is empty.
            self.comb += self.rmw_rgrant.eq(~r_buffer_queue & (r_buffer_level == 0))

            # Prevent new read on Read-Modify-Write request.
            self.comb += If(self.rmw_request,
                r_buffer_queue.eq(0),
                can_read.eq(0)
            )

            # Disconnect regular Datapath on a Read-Modify-Write cycle.
            self.comb += If(self.rmw_request & self.rmw_rgrant,
                port.rdata.ready.eq(1),
                r_buffer.sink.valid.eq(0)
            )

# LiteDRAMAXI2Native -------------------------------------------------------------------------------

class LiteDRAMAXI2Native(Module):
    def __init__(self, axi, port, w_buffer_depth=16, r_buffer_depth=16, base_address=0x00000000, with_read_modify_write=False):

        # # #

        # Write path -------------------------------------------------------------------------------
        self.submodules.write = LiteDRAMAXI2NativeW(axi, port, w_buffer_depth, base_address, with_read_modify_write)

        # Read path --------------------------------------------------------------------------------
        self.submodules.read = LiteDRAMAXI2NativeR(axi, port, r_buffer_depth, base_address, with_read_modify_write)

        # Write / Read arbitration -----------------------------------------------------------------
        arbiter = RoundRobin(2, SP_CE)
        self.submodules += arbiter
        self.comb += arbiter.ce.eq(~port.cmd.valid | (port.cmd.ready & port.cmd.last))
        for i, master in enumerate([self.write, self.read]):
            self.comb += arbiter.request[i].eq(master.cmd_request)
            self.comb += master.cmd_grant.eq(arbiter.grant == i)

        # Read-Modify-Write ------------------------------------------------------------------------
        if with_read_modify_write:
            self.comb += [
                # Connect RMW-Request/Grant between Write and Read paths.
                self.read.rmw_request.eq(self.write.rmw_request),
                self.write.rmw_rgrant.eq(self.read.rmw_rgrant),
            ]



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import os
import argparse

from migen import *

from litex.build.generic_platform import Pins, Subsignal
from litex.build.sim.config import SimConfig

from litex.soc.interconnect.csr import CSR
from litex.soc.integration.soc_core import SoCCore, soc_core_args, soc_core_argdict
from litex.soc.integration.builder import builder_args, builder_argdict, Builder
from litex.soc.cores.cpu import CPUS

from litedram.gen import LiteDRAMCoreControl
from litedram.modules import SDRAMModule, _TechnologyTimings, _SpeedgradeTimings
from litedram.core.controller import ControllerSettings
from litedram.phy.model import DFITimingsChecker, _speedgrade_timings, _technology_timings
from litedram.phy.dfi import DFIRateConverter

from litedram.phy.sim_utils import Clocks, CRG, Platform
from litedram.phy.lpddr5.simphy import LPDDR5SimPHY
from litedram.phy.lpddr5.sim import LPDDR5Sim

# Platform -----------------------------------------------------------------------------------------

_io = [
    # clocks added in main
    ("lpddr5", 0,
        Subsignal("ck",      Pins(1)),  # ck_t/ck_c
        Subsignal("cs",      Pins(1)),
        Subsignal("ca",      Pins(7)),
        Subsignal("dq",      Pins(16)),
        Subsignal("wck",     Pins(2)),  # wck[1:0]_t/wck[1:0]_c
        Subsignal("rdqs",    Pins(2)),  # rdqs[1:0]_t/rdqs[1:0]_c
        Subsignal("dmi",     Pins(2)),
        Subsignal("reset_n", Pins(1)),
    ),
]

# Clocks -------------------------------------------------------------------------------------------

def get_clocks(sys_clk_freq, wck_ck_ratio, dfi_converter_ratio):
    # sys is the main system clock
    # DFI rate converter moves PHY to dfi_converter_ratio*sys
    # PHY CK = dfi_converter_ratio*sys
    # PHY WCK = wck_ck_ratio*dfi_converter_ratio*sys
    # DDR is simulated using twice the clock, this is the case for CA signal on CK and all WCK signals
    defs = {
        "phy": dfi_converter_ratio,
        "ck": dfi_converter_ratio,
        "ck_ddr": 2*dfi_converter_ratio,
        "ca_ddr": 4*dfi_converter_ratio,
        "wck_ddr": 2*wck_ck_ratio*dfi_converter_ratio,
    }
    clocks = {"sys": dict(freq_hz=sys_clk_freq)}
    for name, ratio in defs.items():
        clocks[f"sys{ratio}x"] = dict(freq_hz=ratio*sys_clk_freq)
        clocks[f"sys{ratio}x_180"] = dict(freq_hz=ratio*sys_clk_freq, phase_deg=180)
    return Clocks(clocks)

# SoC ----------------------------------------------------------------------------------------------

class LPDDR5ExampleModule(SDRAMModule):
    # 16B mode, 8Gb (32Mb x 16DQ x 16 banks)
    # x16, DVFSC disabled, Write Link ECC diabled
    # TODO: missing timings: tCCD, tZQCS
    memtype = "LPDDR5"

    nbanks      = 16
    nrows       = 32768
    ncols       = 1024  # specs split this into C[5:0] "column address" and B[3:0] "burst address"

    # TODO: find a way to select if we need masked writes
    tccd = {"write": (8, None), "masked-write": (32, None)}

    technology_timings = _TechnologyTimings(tREFI=32e6/8192, tWTR=(4, 12), tCCD=tccd["masked-write"], tRRD=(2, 5), tZQCS=(128, 80))
    speedgrade_timings = {
        "default": _SpeedgradeTimings(tRP=(2, 21), tRCD=(2, 18), tWR=(3, 34), tRFC=210, tFAW=20, tRAS=(3, 42)),  # TODO: tRAS_max
    }

class SimSoC(SoCCore):
    """Simulation of SoC with LPDDR5 DRAM"""
    def __init__(self, clocks, log_level,
            auto_precharge=False, with_refresh=True, trace_reset=0, disable_delay=False,
            masked_write=True, finish_after_memtest=False, wck_ck_ratio=2, dfi_converter_ratio=1, **kwargs):
        platform     = Platform(_io, clocks)
        sys_clk_freq = clocks["sys"]["freq_hz"]

        # SoCCore ----------------------------------------------------------------------------------
        super().__init__(platform,
            clk_freq    = sys_clk_freq,
            ident       = "LiteX Simulation",
            cpu_variant = "lite",
            **kwargs)

        # CRG --------------------------------------------------------------------------------------
        self.submodules.crg = CRG(platform, clocks)

        # Debugging --------------------------------------------------------------------------------
        platform.add_debug(self, reset=trace_reset)

        # LPDDR5 -----------------------------------------------------------------------------------
        pads = platform.request("lpddr5")
        sdram_module = LPDDR5ExampleModule(sys_clk_freq, f"1:{dfi_converter_ratio}")

        # Map system and Serializer/Deserializer clocks to higher frequency
        clock_mapping = {"sys": f"sys{dfi_converter_ratio}x"}
        for ratio in [2, 2*wck_ck_ratio]:
            clock_mapping[f"sys{ratio}x"] = f"sys{dfi_converter_ratio * ratio}x"
            clock_mapping[f"sys{ratio}x_180"] = f"sys{dfi_converter_ratio * ratio}x_180"

        sim_phy_cls = DFIRateConverter.phy_wrapper(
            phy_cls   = LPDDR5SimPHY,
            ratio     = dfi_converter_ratio,
            phy_attrs = [
                "pads",
                "memtype",
                "nranks",
                "databits",
                "addressbits",
                "bankbits",
                "nphases",
                "twck",
                "ser_latency",
                "des_latency",
                "out",
            ],
            clock_mapping=clock_mapping,
            serdes_reset_cnt=0,
        )

        self.submodules.ddrphy = sim_phy_cls(
            sys_clk_freq       = dfi_converter_ratio*sys_clk_freq,
            aligned_reset_zero = True,
            masked_write       = masked_write,
            wck_ck_ratio       = wck_ck_ratio,
        )

        for p in ["ck", "cs", "ca", "dq", "wck", "rdqs", "dmi", "reset_n"]:
            self.comb += getattr(pads, p).eq(getattr(self.ddrphy.pads, p))

        controller_settings = ControllerSettings()
        controller_settings.auto_precharge = auto_precharge
        controller_settings.with_refresh = with_refresh

        self.add_sdram("sdram",
            phy                     = self.ddrphy,
            module                  = sdram_module,
            origin                  = self.mem_map["main_ram"],
            size                    = kwargs.get("max_sdram_size", 0x40000000),
            l2_cache_size           = kwargs.get("l2_size", 8192),
            l2_cache_min_data_width = kwargs.get("min_l2_data_width", 128),
            l2_cache_reverse        = False,
            controller_settings     = controller_settings
        )
        # Reduce memtest size for simulation speedup
        self.add_constant("MEMTEST_DATA_SIZE", 8*1024)
        self.add_constant("MEMTEST_ADDR_SIZE", 8*1024)

        # LPDDR5 Sim -------------------------------------------------------------------------------
        self.submodules.lpddr5sim = LPDDR5Sim(
            pads          = self.ddrphy.pads,
            ck_freq       = dfi_converter_ratio*sys_clk_freq,
            wck_ck_ratio  = wck_ck_ratio,
            check_timings = not disable_delay,
            log_level     = log_level,
            logger_kwargs = dict(
                clk_freq_cd = f"sys{2*wck_ck_ratio*dfi_converter_ratio}x",
                clk_freq    = 2*wck_ck_ratio*dfi_converter_ratio * sys_clk_freq,
                with_csrs   = True,
            ),
        )

        # Reuse DFITimingsChecker from phy/model.py
        nphases = self.sdram.controller.settings.phy.nphases
        timings = {"tCK": (1e9 / sys_clk_freq) / nphases}
        for name in _speedgrade_timings + _technology_timings:
            timings[name] = sdram_module.get(name)

        self.submodules.dfi_timings_checker = DFITimingsChecker(
            dfi          = self.ddrphy.dfi,
            nbanks       = 2**self.sdram.controller.settings.geom.bankbits,
            nphases      = nphases,
            timings      = timings,
            refresh_mode = sdram_module.timing_settings.fine_refresh_mode,
            memtype      = self.sdram.controller.settings.phy.memtype,
            verbose      = False,
        )

        self.submodules.ddrctrl = LiteDRAMCoreControl()

        self.add_constant("CONFIG_SIM_DISABLE_BIOS_PROMPT")
        if disable_delay:
            self.add_constant("CONFIG_DISABLE_DELAYS")
            # when delays are disabled we may experience timing violations during software control
            self.comb += self.dfi_timings_checker.logging_enabled.eq(self.ddrctrl.init_done.storage)
        if finish_after_memtest:
            self.sync += If(self.ddrctrl.init_done.storage, Finish())

        # Debug info -------------------------------------------------------------------------------
        def dump(obj):
            print()
            print(" " + obj.__class__.__name__)
            print(" " + "-" * len(obj.__class__.__name__))
            d = obj if isinstance(obj, dict) else vars(obj)
            for var, val in d.items():
                if var == "self":
                    continue
                if isinstance(val, Signal):
                    val = "Signal(reset={})".format(val.reset.value)
                print("  {}: {}".format(var, val))

        print("=" * 80)
        dump(clocks)
        if hasattr(self.ddrphy, "phy"):
            dump(self.ddrphy.phy.settings)
        dump(self.ddrphy.settings)
        dump(sdram_module.geom_settings)
        dump(sdram_module.timing_settings)
        print()
        print("=" * 80)

# Build --------------------------------------------------------------------------------------------

def generate_gtkw_savefile(builder, vns, trace_fst):
    from litex.build.sim import gtkwave as gtkw

    dumpfile = os.path.join(builder.gateware_dir, "sim.{}".format("fst" if trace_fst else "vcd"))
    savefile = os.path.join(builder.gateware_dir, "sim.gtkw")
    soc = builder.soc

    with gtkw.GTKWSave(vns, savefile=savefile, dumpfile=dumpfile) as save:
        save.clocks()
        save.add(soc.bus.slaves["main_ram"], mappers=[gtkw.wishbone_sorter(), gtkw.wishbone_colorer()])
        save.fsm_states(soc)
        # all dfi signals
        save.add(soc.ddrphy.dfi, mappers=[gtkw.dfi_sorter(), gtkw.dfi_in_phase_colorer()])
        # each phase in separate group
        with save.gtkw.group("dfi phaseX", closed=True):
            for i, phase in enumerate(soc.ddrphy.dfi.phases):
                save.add(phase, group_name="dfi p{}".format(i), mappers=[
                    gtkw.dfi_sorter(phases=False),
                    gtkw.dfi_in_phase_colorer(),
                ])
        # only dfi command signals
        save.add(soc.ddrphy.dfi, group_name="dfi commands", mappers=[
            gtkw.regex_filter(gtkw.suffixes2re(["cas_n", "ras_n", "we_n"])),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        # only dfi data signals
        save.add(soc.ddrphy.dfi, group_name="dfi wrdata", mappers=[
            gtkw.regex_filter(["wrdata$", "p0.*wrdata_en$"]),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        save.add(soc.ddrphy.dfi, group_name="dfi wrdata_mask", mappers=[
            gtkw.regex_filter(gtkw.suffixes2re(["wrdata_mask"])),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        save.add(soc.ddrphy.dfi, group_name="dfi rddata", mappers=[
            gtkw.regex_filter(gtkw.suffixes2re(["rddata", "p0.*rddata_valid"])),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        # serialization
        out = soc.ddrphy.out
        save.group([out.cs, out.wck[0], out.dq_o[0], out.dq_oe, out.dmi_o[0], out.dmi_oe],
            group_name = "out serialization",
            mappers = [
                gtkw.regex_colorer({
                    "yellow": gtkw.suffixes2re(["cs"]),
                    "orange": ["_o[^e]"],
                    "red": gtkw.suffixes2re(["oe"]),
                })
            ]
        )
        save.group([out.dq_i[0], out.dq_oe, out.rdqs_i[0], out.rdqs_oe],
            group_name = "out deserialization",
            mappers = [
                gtkw.regex_colorer({
                    "red": gtkw.suffixes2re(["oe"]),
                })
            ]
        )
        # dram pads
        save.group([s for s in vars(soc.ddrphy.pads).values() if isinstance(s, Signal)],
            group_name = "pads",
            closed = False,
            mappers = [
                gtkw.regex_filter(["_[io]$"], negate=True),
                gtkw.regex_sorter(gtkw.suffixes2re(["reset_n", "ck", "cs", "ca", "dq", "wck", "dmi", "rdqs"])),
                gtkw.regex_colorer({
                    "yellow": gtkw.suffixes2re(["cs", "ca"]),
                    "orange": gtkw.suffixes2re(["dq", "wck", "dmi"]),
                    "red": gtkw.suffixes2re(["oe"]),
                }),
            ],
        )

        from litedram.phy.lpddr5.sim import gtkw_dbg
        for name in "cmd_info cmd_buf".split():
            save.add(gtkw_dbg[name], group_name=name, closed=False,
                # mappers=[gtkw.endpoint_filter(payload=False)],
                mappers=[gtkw.endpoint_filter()],
            )

def main():
    parser = argparse.ArgumentParser(description="Generic LiteX SoC Simulation")
    builder_args(parser.add_argument_group(title="Builder"))
    soc_core_args(parser.add_argument_group(title="SoC Core"))
    group = parser.add_argument_group(title="LPDDR4 simulation")
    group.add_argument("--sdram-verbosity",      default=0,               help="Set SDRAM checker verbosity")
    group.add_argument("--trace",                action="store_true",     help="Enable Tracing")
    group.add_argument("--trace-fst",            action="store_true",     help="Enable FST tracing (default=VCD)")
    group.add_argument("--trace-start",          default=0,               help="Cycle to start tracing")
    group.add_argument("--trace-end",            default=-1,              help="Cycle to end tracing")
    group.add_argument("--trace-reset",          default=0,               help="Initial traceing state")
    group.add_argument("--sys-clk-freq",         default="100e6",         help="System clock frequency")
    group.add_argument("--auto-precharge",       action="store_true",     help="Use DRAM auto precharge")
    group.add_argument("--no-refresh",           action="store_true",     help="Disable DRAM refresher")
    group.add_argument("--wck-ck-ratio", default=2, type=int, choices=[2, 4], help="WCK:CK ratio")
    group.add_argument("--dfi-converter-ratio", default=1, type=int,      help="Conversion ratio between sys clock and PHY CK")
    group.add_argument("--log-level",            default="all=INFO",      help="Set simulation logging level")
    group.add_argument("--disable-delay",        action="store_true",     help="Disable CPU delays")
    group.add_argument("--gtkw-savefile",        action="store_true",     help="Generate GTKWave savefile")
    group.add_argument("--no-masked-write",      action="store_true",     help="Use LPDDR4 WRITE instead of MASKED-WRITE")
    group.add_argument("--no-run",               action="store_true",     help="Don't run the simulation, just generate files")
    group.add_argument("--finish-after-memtest", action="store_true",     help="Stop simulation after DRAM memory test")
    args = parser.parse_args()

    soc_kwargs     = soc_core_argdict(args)
    builder_kwargs = builder_argdict(args)

    sim_config = SimConfig()
    sys_clk_freq = int(float(args.sys_clk_freq))
    clocks = get_clocks(sys_clk_freq, wck_ck_ratio=args.wck_ck_ratio, dfi_converter_ratio=args.dfi_converter_ratio)
    clocks.add_clockers(sim_config)

    # Configuration --------------------------------------------------------------------------------
    if soc_kwargs["uart_name"] == "serial":
        soc_kwargs["uart_name"] = "sim"
        sim_config.add_module("serial2console", "serial")
    args.with_sdram = True
    soc_kwargs["integrated_main_ram_size"] = 0x0
    soc_kwargs["sdram_verbosity"]          = int(args.sdram_verbosity)

    # SoC ------------------------------------------------------------------------------------------
    soc = SimSoC(
        clocks          = clocks,
        auto_precharge  = args.auto_precharge,
        with_refresh    = not args.no_refresh,
        trace_reset     = int(args.trace_reset),
        log_level       = args.log_level,
        disable_delay   = args.disable_delay,
        masked_write    = not args.no_masked_write,
        finish_after_memtest = args.finish_after_memtest,
        wck_ck_ratio    = args.wck_ck_ratio,
        dfi_converter_ratio = args.dfi_converter_ratio,
        **soc_kwargs)

    # Build/Run ------------------------------------------------------------------------------------
    builder_kwargs["csr_csv"] = "csr.csv"
    builder = Builder(soc, **builder_kwargs)
    build_kwargs = dict(
        sim_config  = sim_config,
        trace       = args.trace,
        trace_fst   = args.trace_fst,
        trace_start = int(args.trace_start),
        trace_end   = int(args.trace_end)
    )
    vns = builder.build(run=False, **build_kwargs)

    if args.gtkw_savefile:
        generate_gtkw_savefile(builder, vns, trace_fst=args.trace_fst)

    if not args.no_run:
        builder.build(build=False, **build_kwargs)

if __name__ == "__main__":
    main()
